#  # noqa: N999
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""Main functionality to run a pelicun calculation from the command line."""

from __future__ import annotations

import argparse
import json
import os
import sys
from pathlib import Path
from time import gmtime, strftime
from typing import Hashable

import colorama
import jsonschema
import numpy as np
import pandas as pd
from colorama import Fore, Style
from jsonschema import validate

from pelicun import base
from pelicun.assessment import DLCalculationAssessment
from pelicun.auto import auto_populate
from pelicun.base import (
    convert_to_MultiIndex,
    convert_to_SimpleIndex,
    describe,
    get,
    is_specified,
    is_unspecified,
    str2bool,
    update,
    update_vals,
)
from pelicun.file_io import substitute_default_path
from pelicun.pelicun_warnings import PelicunInvalidConfigError

colorama.init()
sys.path.insert(0, Path(__file__).resolve().parent.absolute().as_posix())


def log_msg(msg: str, color_codes: tuple[str, str] | None = None) -> None:
    """
    Print a formatted log message with a timestamp.

    Parameters
    ----------
    msg : str
        The message to print.
    color_codes : tuple, optional
        Color codes for formatting the message. Default is None.

    """
    if color_codes:
        cpref, csuff = color_codes
        print(  # noqa: T201
            f'{strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())} '
            f'{cpref}'
            f'{msg}'
            f'{csuff}'
        )
    else:
        print(f'{strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())} {msg}')  # noqa: T201


# list of output files help perform safe initialization of output dir
known_output_files = [
    'DEM_sample.zip',
    'DEM_stats.csv',
    'CMP_sample.zip',
    'CMP_stats.csv',
    'DMG_sample.zip',
    'DMG_stats.csv',
    'DMG_grp.zip',
    'DMG_grp_stats.csv',
    'DV_repair_sample.zip',
    'DV_repair_stats.csv',
    'DV_repair_grp.zip',
    'DV_repair_grp_stats.csv',
    'DV_repair_agg.zip',
    'DV_repair_agg_stats.csv',
    'DL_summary.csv',
    'DL_summary_stats.csv',
]

full_out_config = {
    'Demand': {'Sample': True, 'Statistics': True},
    'Asset': {'Sample': True, 'Statistics': True},
    'Damage': {
        'Sample': True,
        'Statistics': True,
        'GroupedSample': True,
        'GroupedStatistics': True,
    },
    'Loss': {
        'Repair': {
            'Sample': True,
            'Statistics': True,
            'GroupedSample': True,
            'GroupedStatistics': True,
            'AggregateSample': True,
            'AggregateStatistics': True,
        }
    },
    'Format': {'CSV': True, 'JSON': True},
}

regional_out_config = {
    'Demand': {'Sample': True, 'Statistics': False},
    'Asset': {'Sample': True, 'Statistics': False},
    'Damage': {
        'Sample': False,
        'Statistics': False,
        'GroupedSample': True,
        'GroupedStatistics': True,
    },
    'Loss': {
        'Repair': {
            'Sample': True,
            'Statistics': True,
            'GroupedSample': True,
            'GroupedStatistics': True,
            'AggregateSample': True,
            'AggregateStatistics': True,
        }
    },
    'Format': {'CSV': False, 'JSON': True},
    'Settings': {
        'CondenseDS': True,
        'SimpleIndexInJSON': True,
        'AggregateColocatedComponentResults': True,
    },
}

pbe_settings = {
    'CondenseDS': False,
    'SimpleIndexInJSON': False,
    'AggregateColocatedComponentResults': True,
}


def convert_df_to_dict(data: pd.DataFrame | pd.Series, axis: int = 1) -> dict:
    """
    Convert a pandas DataFrame to a dictionary.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to be converted.
    axis : int, optional
        The axis to consider for the conversion.
        * If 1 (default), the DataFrame is used as-is.
        * If 0, the DataFrame is transposed before conversion.

    Returns
    -------
    dict
        A dictionary representation of the DataFrame. The structure of
        the dictionary depends on the levels in the DataFrame's
        MultiIndex columns.

    Raises
    ------
    ValueError
        If the axis is not 0 or 1.

    Notes
    -----
    * If the columns have multiple levels, the function will
      recursively convert sub-DataFrames.
    * If the column labels at any level are numeric, they will be
      converted to a list of floats.
    * If the column labels are non-numeric, a dictionary will be
      created with the index labels as keys and the corresponding data
      as values.

    """
    out_dict: dict[Hashable, object] = {}

    if axis == 1:
        df_in = data
    elif axis == 0:
        df_in = data.T
    else:
        msg = '`axis` must be `0` or `1`'
        raise ValueError(msg)

    multiindex = df_in.columns

    for label in multiindex.unique(level=0):
        out_dict.update({label: np.nan})

        sub_df = df_in[label]

        skip_sub = True

        if multiindex.nlevels > 1:
            skip_sub = False

            if isinstance(sub_df, pd.Series) or (
                (len(sub_df.columns) == 1) and (sub_df.columns[0] == '')  # noqa: PLC1901
            ):
                skip_sub = True

            if not skip_sub:
                out_dict[label] = convert_df_to_dict(sub_df)

        if skip_sub:
            if np.all(sub_df.index.astype(str).str.isnumeric()):
                out_dict_label = df_in[label].astype(float)
                out_dict[label] = out_dict_label.tolist()
            else:
                out_dict[label] = {key: sub_df.loc[key] for key in sub_df.index}

    return out_dict


def run_pelicun(  # noqa: C901
    config_path: str,
    demand_file: str,
    output_path: str | None,
    realizations: int,
    auto_script_path: str | None,
    custom_model_dir: str | None,
    output_format: list | None,
    *,
    detailed_results: bool,
    coupled_edp: bool,
) -> None:
    """
    Use settings in the config JSON to prepare and run a Pelicun calculation.

    Parameters
    ----------
    config_path: string
        Path pointing to the location of the JSON configuration file.
    demand_file: string
        Path pointing to the location of a CSV file with the demand data.
    output_path: string, optional
        Path pointing to the location where results shall be saved.
    realizations: int, optional
        Number of realizations to generate.
    auto_script_path: string, optional
        Path pointing to the location of a Python script with an auto_populate
        method that automatically creates the performance model using data
        provided in the AIM JSON file.
    custom_model_dir: string, optional
        Path pointing to a directory with files that define user-provided model
        parameters for a customized damage and loss assessment.
    output_format: list, optional.
        Type of output format, JSON or CSV.
        Valid options: ['csv', 'json'], ['csv'], ['json'], [], None
    detailed_results: bool, optional
        If False, only the main statistics are saved.
    coupled_edp: bool, optional
        If True, EDPs are not resampled and processed in order.

    """
    # Initial setup -----------------------------------------------------------

    # get the absolute path to the config file
    config_path_p = Path(config_path).resolve()

    # If the output path was not specified, results are saved in the
    # directory of the input file.
    if output_path is None:
        output_path_p = config_path_p.parents[0]
    else:
        output_path_p = Path(output_path).resolve()
    # create the directory if it does not exist
    if not output_path_p.exists():
        output_path_p.mkdir(parents=True)

    # parse the config file
    config = _parse_config_file(
        config_path_p,
        output_path_p,
        Path(auto_script_path).resolve() if auto_script_path is not None else None,
        demand_file,
        realizations,
        output_format,
        custom_model_dir,
        coupled_edp=coupled_edp,
        detailed_results=detailed_results,
    )

    # An undefined config means that we do not need to run a simulation
    # Such config is not an error, those are caught during parsing. This is the
    # result of an intentional no-simulation request during auto-population.
    if config is None:
        return

    # List to keep track of the generated output files.
    out_files: list[str] = []

    _remove_existing_files(output_path_p, known_output_files)

    # Run the assessment
    assessment = DLCalculationAssessment(config_options=get(config, 'DL/Options'))

    assessment.calculate_demand(
        demand_path=Path(get(config, 'DL/Demands/DemandFilePath')).resolve(),
        collapse_limits=get(config, 'DL/Demands/CollapseLimits', default=None),
        length_unit=get(config, 'GeneralInformation/units/length', default=None),
        demand_calibration=get(config, 'DL/Demands/Calibration', default=None),
        sample_size=get(config, 'DL/Options/Sampling/SampleSize'),
        demand_cloning=get(config, 'DL/Demands/DemandCloning', default=None),
        residual_drift_inference=get(
            config, 'DL/Demands/InferResidualDrift', default=None
        ),
        coupled_demands=get(config, 'DL/Demands/CoupledDemands', default=False),
    )

    if is_specified(config, 'DL/Asset'):
        assessment.calculate_asset(
            num_stories=get(config, 'DL/Asset/NumberOfStories', default=None),
            component_assignment_file=get(
                config, 'DL/Asset/ComponentAssignmentFile', default=None
            ),
            collapse_fragility_demand_type=get(
                config, 'DL/Damage/CollapseFragility/DemandType', default=None
            ),
            component_sample_file=get(
                config, 'DL/Asset/ComponentSampleFile', default=None
            ),
            add_irreparable_damage_columns=get(
                config, 'DL/Damage/IrreparableDamage', default=False
            ),
        )

    if is_specified(config, 'DL/Damage'):
        assessment.calculate_damage(
            length_unit=get(config, 'GeneralInformation/units/length'),
            component_database=get(config, 'DL/Asset/ComponentDatabase'),
            component_database_path=get(
                config, 'DL/Asset/ComponentDatabasePath', default=None
            ),
            collapse_fragility=get(
                config, 'DL/Damage/CollapseFragility', default=None
            ),
            irreparable_damage=get(
                config, 'DL/Damage/IrreparableDamage', default=None
            ),
            damage_process_approach=get(
                config, 'DL/Damage/DamageProcess', default=None
            ),
            damage_process_file_path=get(
                config, 'DL/Damage/DamageProcessFilePath', default=None
            ),
            custom_model_dir=custom_model_dir,
            scaling_specification=get(config, 'DL/Damage/ScalingSpecification'),
            is_for_water_network_assessment='Water'
            in get(config, 'DL/Asset/ComponentDatabase', ''),
        )

    if is_unspecified(config, 'DL/Losses/Repair'):
        agg_repair = None
    else:
        # Currently we only support `Repair` consequences.
        # We will need to make changes here when we start to include
        # more consequences.

        agg_repair, _ = assessment.calculate_loss(
            loss_map_approach=get(config, 'DL/Losses/Repair/MapApproach'),
            occupancy_type=get(config, 'DL/Asset/OccupancyType'),
            consequence_database=get(config, 'DL/Losses/Repair/ConsequenceDatabase'),
            consequence_database_path=get(
                config, 'DL/Losses/Repair/ConsequenceDatabasePath'
            ),
            custom_model_dir=custom_model_dir,
            damage_process_approach=get(
                config, 'DL/Damage/DamageProcess', default='User Defined'
            ),
            replacement_cost_parameters=get(
                config, 'DL/Losses/Repair/ReplacementCost'
            ),
            replacement_time_parameters=get(
                config, 'DL/Losses/Repair/ReplacementTime'
            ),
            replacement_carbon_parameters=get(
                config, 'DL/Losses/Repair/ReplacementCarbon'
            ),
            replacement_energy_parameters=get(
                config, 'DL/Losses/Repair/ReplacementEnergy'
            ),
            loss_map_path=get(config, 'DL/Losses/Repair/MapFilePath'),
            decision_variables=_parse_decision_variables(config),
            replacement_configuration=None,  # will be used later
            loss_combination_method=get(
                config, 'DL/Losses/Repair/CombinationMethod'
            ),
        )

    summary, summary_stats = _result_summary(assessment, agg_repair)

    # Save the results into files

    if is_specified(config, 'DL/Outputs/Demand'):
        output_config = get(config, 'DL/Outputs/Demand')
        _demand_save(output_config, assessment, output_path_p, out_files)

    if is_specified(config, 'DL/Outputs/Asset'):
        output_config = get(config, 'DL/Outputs/Asset')
        _asset_save(
            output_config,
            assessment,
            output_path_p,
            out_files,
            aggregate_colocated=get(
                config,
                'DL/Outputs/Settings/AggregateColocatedComponentResults',
                default=False,
            ),
        )

    if is_specified(config, 'DL/Outputs/Damage'):
        output_config = get(config, 'DL/Outputs/Damage')
        _damage_save(
            output_config,
            assessment,
            output_path_p,
            out_files,
            aggregate_colocated=get(
                config,
                'DL/Outputs/Settings/AggregateColocatedComponentResults',
                default=False,
            ),
            condense_ds=get(
                config,
                'DL/Outputs/Settings/CondenseDS',
                default=False,
            ),
        )

    if is_specified(config, 'DL/Outputs/Loss/Repair'):
        output_config = get(config, 'DL/Outputs/Loss/Repair')
        assert agg_repair is not None
        _loss_save(
            output_config,
            assessment,
            output_path_p,
            out_files,
            agg_repair,
            aggregate_colocated=get(
                config,
                'DL/Outputs/Settings/AggregateColocatedComponentResults',
                default=False,
            ),
        )
    _summary_save(summary, summary_stats, output_path_p, out_files)
    _create_json_files_if_requested(config, out_files, output_path_p)
    _remove_csv_files_if_not_requested(config, out_files, output_path_p)


def _parse_decision_variables(config: dict) -> tuple[str, ...]:
    """
    Parse decision variables from the config file.

    Parameters
    ----------
    config : dict
        The configuration dictionary.

    Returns
    -------
    list
        List of decision variables.

    """
    decision_variables: list[str] = []
    if get(config, 'DL/Losses/Repair/DecisionVariables', default=False) is not False:
        for dv_i, dv_status in get(
            config, 'DL/Losses/Repair/DecisionVariables'
        ).items():
            if dv_status is True:
                decision_variables.append(dv_i)
    return tuple(decision_variables)


def _remove_csv_files_if_not_requested(
    config: dict, out_files: list[str], output_path: Path
) -> None:
    """
    Remove CSV files if not requested in config.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    out_files : list
        List of output file names.
    output_path : Path
        Path to the output directory.
    """
    # Don't proceed if CSV files were requested.
    if get(config, 'DL/Outputs/Format/CSV', default=False) is True:
        return

    for filename in out_files:
        # keep the DL_summary and DL_summary_stats files
        if 'DL_summary' in filename:
            continue
        Path(output_path / filename).unlink()


def _summary_save(
    summary: pd.DataFrame,
    summary_stats: pd.DataFrame,
    output_path: Path,
    out_files: list[str],
) -> None:
    """
    Save summary results to CSV files.

    Parameters
    ----------
    summary : pd.DataFrame
        Summary DataFrame.
    summary_stats : pd.DataFrame
        Summary statistics DataFrame.
    output_path : Path
        Path to the output directory.
    out_files : list
        List of output file names.

    """
    # save summary sample
    if summary is not None:
        summary.to_csv(output_path / 'DL_summary.csv', index_label='#')
        out_files.append('DL_summary.csv')

    # save summary statistics
    if summary_stats is not None:
        summary_stats.to_csv(output_path / 'DL_summary_stats.csv')
        out_files.append('DL_summary_stats.csv')


def _parse_config_file(  # noqa: C901, PLR0912
    config_path: Path,
    output_path: Path,
    auto_script_path: Path | None,
    demand_file: str,
    realizations: int,
    output_format: list | None,
    custom_model_dir: str | None,
    *,
    coupled_edp: bool,
    detailed_results: bool,
) -> dict[str, object] | None:
    """
    Parse and validate the config file for Pelicun.

    Parameters
    ----------
    config_path : str
        Path to the configuration file.
    output_path : Path
        Directory for output files.
    auto_script_path : Path
        Path to the auto-generation script.
    demand_file : str
        Path to the demand data file.
    realizations : int
        Number of realizations.
    output_format : str
        Output format (CSV, JSON).
    custom_model_dir: str, optional
        String pointing to a directory with files that define user-provided model
        parameters for a customized damage and loss assessment.
    coupled_EDP : bool
        Whether to consider coupled EDPs.
    detailed_results : bool
        Whether to generate detailed results.

    Returns
    -------
    dict
        Parsed and validated configuration.

    Raises
    ------
    PelicunInvalidConfigError
      If the provided config file does not conform to the schema or
      there are issues with the specified values.

    """
    # open the config file and parse it
    with Path(config_path).open(encoding='utf-8') as f:
        config = json.load(f)

    # load the schema
    with Path(f'{base.pelicun_path}/settings/input_schema.json').open(
        encoding='utf-8'
    ) as f:
        schema = json.load(f)

    # add the demand file to the DL if needed
    if is_specified(config, 'DL'):
        if is_unspecified(config, 'DL/Demands/DemandFilePath'):
            update(config, '/DL/Demands/DemandFilePath', demand_file)

    # Validate the configuration against the schema
    try:
        validate(instance=config, schema=schema)
    except jsonschema.exceptions.ValidationError as exc:
        msg = 'The provided config file does not conform to the schema.'
        raise PelicunInvalidConfigError(msg) from exc

    # identify the folder with the damage and loss model data
    if is_unspecified(config, 'DL'):
        dl_method = get(config, 'Applications/DL/ApplicationData/DL_Method')

        if dl_method == 'User-provided Models':
            if custom_model_dir is not None:
                dl_model_folder = custom_model_dir
            else:
                dl_model_folder = get(
                    config, 'Applications/DL/ApplicationData/custom_model_dir'
                )

            assert isinstance(dl_model_folder, str)
            dl_model_folder = Path(dl_model_folder).resolve()
            assert dl_model_folder.exists(), f'{dl_model_folder} does not exist'
            assert dl_model_folder.is_dir(), f'{dl_model_folder} is not a directory'

            auto_script_paths = [
                Path(dl_model_folder / 'pelicun_config.py').resolve()
            ]

        else:
            dl_methods = [m.strip() for m in dl_method.split(',')]

            auto_script_paths = []
            for dl_method in dl_methods:
                auto_script_path = substitute_default_path(
                    [f'PelicunDefault/{dl_method}/pelicun_config.py']
                )[0]
                auto_script_paths.append(Path(auto_script_path).resolve())

        for auto_script_path in auto_script_paths:
            if not auto_script_path.exists():
                msg = (
                    f'No `DL` entry in config file and the following path '
                    f'does not point to a valid pelicun configuration file: '
                    f'{auto_script_path}.'
                )
                raise PelicunInvalidConfigError(msg)

        # Add the demandFile to the config dict to allow demand-dependent auto-population
        update(config, '/DL/Demands/DemandFilePath', demand_file)
        update(config, '/DL/Demands/SampleSize', str(realizations))

        for script_id, auto_script_path in enumerate(auto_script_paths):
            log_msg(f'Configuring Pelicun using {auto_script_path}')

            if script_id == 0:
                config_ap, comp = auto_populate(config, auto_script_path, script_id)

            else:
                config_ap_i, comp_i = auto_populate(
                    config, auto_script_path, script_id
                )

                comp = pd.concat([comp, comp_i])

                # TODO(AZS): Currently, this is set up to work with the old flood rules
                # Requires updating once the inference is moved to BRAILS and the flood
                # config is updated.
                # Requires further updating to make it more generic and support more than
                # just hurricane wind & surge
                update(
                    config_ap,
                    'DL/Asset/ComponentDatabase',
                    (
                        f"{get(config_ap, 'DL/Asset/ComponentDatabase')},"
                        f"{get(config_ap_i, 'DL/Asset/ComponentDatabase')}"
                    ),
                )

                update(
                    config_ap,
                    'DL/Losses/Repair/ConsequenceDatabase',
                    (
                        f"{get(config_ap, 'DL/Losses/Repair/ConsequenceDatabase')},"
                        f"{get(config_ap_i, 'DL/Losses/Repair/ConsequenceDatabase')}"
                    ),
                )

                update(
                    config_ap,
                    'DL/Losses/Repair/CombinationMethod',
                    'Hazus Hurricane',
                )

        if is_unspecified(config_ap, 'DL'):
            msg = (
                'No `DL` entry in config file, and '
                'the prescribed auto-population script failed to identify '
                'a valid damage and loss configuration for this asset. '
            )
            raise PelicunInvalidConfigError(msg)

        if get(config_ap, 'DL') == 'N/A':
            msg = (
                'N/A `DL` entry in config file interpreted as a request to '
                'skip damage and loss simulation for this asset.'
            )
            log_msg(msg)
            return None

        # look for possibly specified assessment options
        try:
            assessment_options = config['Applications']['DL']['ApplicationData'][
                'Options'
            ]
        except KeyError:
            assessment_options = None

        if assessment_options:
            # extend options defined via the auto-population script to
            # include those in the original `config`
            config_ap['Applications']['DL']['ApplicationData'].pop('Options')
            update_vals(
                config_ap['DL']['Options'],
                assessment_options,
                "config_ap['DL']['Options']",
                'assessment_options',
            )

        # add the demand information
        update(config_ap, '/DL/Demands/DemandFilePath', demand_file)
        update(config_ap, '/DL/Demands/SampleSize', str(realizations))

        if coupled_edp is True:
            update(config_ap, 'DL/Demands/CoupledDemands', value=True)

        else:
            update(
                config_ap,
                'DL/Demands/Calibration',
                {'ALL': {'DistributionFamily': 'lognormal'}},
            )

        # save the component data
        comp.to_csv(output_path / 'CMP_QNT.csv')

        # update the config file with the location
        update(
            config_ap,
            'DL/Asset/ComponentAssignmentFile',
            str(output_path / 'CMP_QNT.csv'),
        )

        # if detailed results are not requested, add a lean output config
        if detailed_results is False:
            update(config_ap, 'DL/Outputs', regional_out_config)
        else:
            update(config_ap, 'DL/Outputs', full_out_config)
            # add output settings from regional output config
            if is_unspecified(config_ap, 'DL/Outputs/Settings'):
                update(config_ap, 'DL/Outputs/Settings', {})

            config_ap['DL']['Outputs']['Settings'].update(
                regional_out_config['Settings']
            )

        # if no loss simulation is requested, remove the corresponding outputs
        if is_unspecified(config_ap, 'DL/Losses'):
            update(config_ap, 'DL/Outputs/Loss', {})

        if is_specified(config, 'outputs'):
            if (not config['outputs']['IM']) and (not config['outputs']['EDP']):
                update(config_ap, 'DL/Outputs/Demand', {})
            if not config['outputs']['DM']:
                update(config_ap, 'DL/Outputs/Damage', {})
            if not config['outputs']['DV']:
                update(config_ap, 'DL/Outputs/Loss', {})

        # save the extended config to a file
        config_ap_path = Path(config_path.stem + '_ap.json').resolve()

        with Path(config_ap_path).open('w', encoding='utf-8') as f:
            json.dump(config_ap, f, indent=2)

        update(config, 'DL', get(config_ap, 'DL'))

    # sample size
    sample_size_str = get(config, 'DL/Options/Sampling/SampleSize')
    if not sample_size_str:
        sample_size_str = get(config, 'DL/Demands/SampleSize')
        if not sample_size_str:
            msg = 'Sample size not provided in config file.'
            raise PelicunInvalidConfigError(msg)
    update(config, 'DL/Options/Sampling/SampleSize', int(sample_size_str))

    # provide all outputs if the files are not specified
    if is_unspecified(config, 'DL/Outputs'):
        update(config, 'DL/Outputs', full_out_config)

    # provide outputs in CSV by default
    if is_unspecified(config, 'DL/Outputs/Format'):
        update(config, 'DL/Outputs/Format', {'CSV': True, 'JSON': False})

    # override file format specification if the output_format is
    # provided
    if output_format is not None:
        update(
            config,
            'DL/Outputs/Format',
            {
                'CSV': 'csv' in output_format,
                'JSON': 'json' in output_format,
            },
        )

    # add empty Settings to output config to simplify code below
    if is_unspecified(config, 'DL/Outputs/Settings'):
        update(config, 'DL/Outputs/Settings', pbe_settings)

    if is_unspecified(config, 'DL/Demands'):
        msg = 'Demand configuration missing.'
        raise PelicunInvalidConfigError(msg)

    if is_unspecified(config, 'DL/Asset'):
        msg = 'Asset configuration missing.'
        raise PelicunInvalidConfigError(msg)

    update(
        config,
        'DL/Options/LogFile',
        'pelicun_log.txt',
        only_if_empty_or_none=True,
    )
    update(
        config,
        'DL/Options/Verbose',
        value=True,
        only_if_empty_or_none=True,
    )

    # if the user did not prescribe anything for ListAllDamageStates,
    # then use True as default for DL_calculations regardless of what
    # the Pelicun default is.
    update(
        config,
        'DL/Options/ListAllDamageStates',
        value=True,
        only_if_empty_or_none=True,
    )

    # if the demand file location is not specified in the config file
    # assume there is a `response.csv` file next to the config file.
    update(
        config,
        'DL/Demands/DemandFilePath',
        config_path.parent / 'response.csv',
        only_if_empty_or_none=True,
    )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # backwards-compatibility for v3.2 and earlier | remove after v4.0
    if get(config, 'DL/Losses/BldgRepair', default=False):
        update(config, 'DL/Losses/Repair', get(config, 'DL/Losses/BldgRepair'))
    if get(config, 'DL/Outputs/Loss/BldgRepair', default=False):
        update(
            config,
            'DL/Outputs/Loss/Repair',
            get(config, 'DL/Outputs/Loss/BldgRepair'),
        )
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Cast NumberOfStories to int
    if is_specified(config, 'DL/Asset/NumberOfStories'):
        update(
            config,
            'DL/Asset/NumberOfStories',
            int(get(config, 'DL/Asset/NumberOfStories')),
        )

    # Ensure `DL/Demands/InferResidualDrift` contains a `method`
    if is_specified(config, 'DL/Demands/InferResidualDrift') and is_unspecified(
        config, 'DL/Demands/InferResidualDrift/method'
    ):
        msg = 'No method is specified in residual drift inference configuration.'
        raise PelicunInvalidConfigError(msg)

    # Ensure `DL/Damage/CollapseFragility` contains all required keys.
    if is_specified(config, 'DL/Damage/CollapseFragility'):
        if is_unspecified(
            config, 'DL/Damage/CollapseFragility/CapacityDistribution'
        ):
            config['DL']['Damage']['CollapseFragility']['CapacityDistribution'] = (
                'deterministic'
            )
            config['DL']['Damage']['CollapseFragility']['Theta_1'] = 'N/A'

        for thing in ('CapacityDistribution', 'CapacityMedian', 'Theta_1'):
            if is_unspecified(config, f'DL/Damage/CollapseFragility/{thing}'):
                msg = (
                    f'`{thing}` is missing from DL/Damage/CollapseFragility'
                    f' in the configuration file.'
                )
                raise PelicunInvalidConfigError(msg)

    # Ensure `DL/Damage/IrreparableDamage` contains all required keys.
    if is_specified(config, 'DL/Damage/IrreparableDamage'):
        for thing in ('DriftCapacityMedian', 'DriftCapacityLogStd'):
            if is_unspecified(config, f'DL/Damage/IrreparableDamage/{thing}'):
                msg = (
                    f'`{thing}` is missing from DL/Damage/IrreparableDamage'
                    f' in the configuration file.'
                )
                raise PelicunInvalidConfigError(msg)

    # If the damage process approach is `User Defined` there needs to
    # be a damage process file path.
    if get(config, 'DL/Damage/DamageProcess') == 'User Defined' and is_unspecified(
        config, 'DL/Damage/DamageProcessFilePath'
    ):
        msg = (
            'When `DL/Damage/DamageProcess` is set to `User Defined`, '
            'a path needs to be specified under '
            '`DL/Damage/DamageProcessFilePath`.'
        )
        raise PelicunInvalidConfigError(msg)

    # Getting results requires running the calculations.
    if is_specified(config, 'DL/Outputs/Asset') and is_unspecified(
        config, 'DL/Asset'
    ):
        msg = (
            'No asset data specified in config file. '
            'Cannot generate asset model outputs.'
        )
        raise PelicunInvalidConfigError(msg)

    if is_specified(config, 'DL/Outputs/Damage') and is_unspecified(
        config, 'DL/Damage'
    ):
        msg = (
            'No damage data specified in config file. '
            'Cannot generate damage model outputs.'
        )
        raise PelicunInvalidConfigError(msg)

    if is_specified(config, 'DL/Outputs/Loss') and is_unspecified(
        config, 'DL/Losses'
    ):
        msg = (
            'No loss data specified in config file. '
            'Cannot generate loss model outputs.'
        )
        raise PelicunInvalidConfigError(msg)

    # Ensure only one of `component_assignment_file` or
    # `component_sample_file` is provided.
    if is_specified(config, 'DL/Asset'):
        if (
            (get(config, 'DL/Asset/ComponentAssignmentFile') is None)
            and (get(config, 'DL/Asset/ComponentSampleFile') is None)
        ) or (
            (get(config, 'DL/Asset/ComponentAssignmentFile') is not None)
            and (get(config, 'DL/Asset/ComponentSampleFile') is not None)
        ):
            msg = (
                'In the asset model configuration, it is '
                'required to specify one of `component_assignment_file` '
                'or `component_sample_file`, but not both.'
            )
            raise PelicunInvalidConfigError(msg)

    return config


def _create_json_files_if_requested(
    config: dict, out_files: list[str], output_path: Path
) -> None:
    """
    Create JSON files if requested in the config.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    out_files : list
        List of output file names.
    output_path : Path
        Path to the output directory.

    """
    # If not requested, simply return
    if get(config, 'DL/Outputs/Format/JSON', default=False) is False:
        return

    for filename in out_files:
        filename_json = filename[:-3] + 'json'

        if (
            get(config, 'DL/Outputs/Settings/SimpleIndexInJSON', default=False)
            is True
        ):
            data = pd.read_csv(output_path / filename, index_col=0)
        else:
            data = convert_to_MultiIndex(
                pd.read_csv(output_path / filename, index_col=0), axis=1
            )

        # Check for units information (case-insensitive)
        units_key = None
        for key in data.index:
            if str(key).lower() == 'units':
                units_key = key
                break

        if units_key is not None:
            df_units = convert_to_SimpleIndex(
                data.loc[units_key, :].to_frame().T,  # type: ignore
                axis=1,
            )

            data = data.drop(units_key, axis=0)

            out_dict = convert_df_to_dict(data)

            out_dict.update(
                {
                    'Units': {
                        col: df_units.loc[units_key, col] for col in df_units.columns
                    }
                }
            )

        else:
            out_dict = convert_df_to_dict(data)

        with Path(output_path / filename_json).open('w', encoding='utf-8') as f:
            json.dump(out_dict, f, indent=2)


def _result_summary(
    assessment: DLCalculationAssessment, agg_repair: pd.DataFrame | None
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Generate a summary of the results.

    Parameters
    ----------
    assessment : AssessmentBase
        The assessment object.
    agg_repair : pd.DataFrame
        Aggregate repair data.

    Returns
    -------
    tuple
        Summary DataFrame and summary statistics DataFrame.

    """
    damage_sample = assessment.damage.save_sample()
    if damage_sample is None and agg_repair is None:
        return pd.DataFrame(), pd.DataFrame()

    if damage_sample is not None:
        assert isinstance(damage_sample, pd.DataFrame)
        damage_sample = damage_sample.groupby(level=['cmp', 'ds'], axis=1).sum()  # type: ignore
        assert isinstance(damage_sample, pd.DataFrame)
        damage_sample_s = convert_to_SimpleIndex(damage_sample, axis=1)

        if 'collapse-1' in damage_sample_s.columns:
            damage_sample_s['collapse'] = damage_sample_s['collapse-1']
        else:
            damage_sample_s['collapse'] = np.zeros(damage_sample_s.shape[0])

        if 'irreparable-1' in damage_sample_s.columns:
            damage_sample_s['irreparable'] = damage_sample_s['irreparable-1']
        else:
            damage_sample_s['irreparable'] = np.zeros(damage_sample_s.shape[0])

        damage_sample_s = damage_sample_s[['collapse', 'irreparable']]

    else:
        damage_sample_s = pd.DataFrame()

    if agg_repair is not None:
        agg_repair_s = convert_to_SimpleIndex(agg_repair, axis=1)

    else:
        agg_repair_s = pd.DataFrame()

    summary = pd.concat([agg_repair_s, damage_sample_s], axis=1)

    summary_stats = describe(summary)

    return summary, summary_stats


def _parse_requested_output_file_names(output_config: dict) -> set[str]:
    """
    Parse the output file names from the output configuration.

    Parameters
    ----------
    output_config : dict
        Configuration for output files.

    Returns
    -------
    set
        Set of requested output file names.

    """
    out_reqs = []
    for out, val in output_config.items():
        if val is True:
            out_reqs.append(out)
    return set(out_reqs)


def _demand_save(
    output_config: dict,
    assessment: DLCalculationAssessment,
    output_path: Path,
    out_files: list[str],
) -> None:
    """
    Save demand results to files based on the output config.

    Parameters
    ----------
    output_config : dict
        Configuration for output files.
    assessment : AssessmentBase
        The assessment object.
    output_path : Path
        Path to the output directory.
    out_files : list
        List of output file names.

    """
    out_reqs = _parse_requested_output_file_names(output_config)

    demand_sample, demand_units_series = assessment.demand.save_sample(
        save_units=True
    )
    assert isinstance(demand_sample, pd.DataFrame)
    assert isinstance(demand_units_series, pd.Series)
    demand_units = demand_units_series.to_frame().T

    if 'Sample' in out_reqs:
        demand_sample_s = pd.concat([demand_sample, demand_units])
        demand_sample_s = convert_to_SimpleIndex(demand_sample_s, axis=1)
        demand_sample_s.to_csv(
            output_path / 'DEM_sample.zip',
            index_label=demand_sample_s.columns.name,
            compression={'method': 'zip', 'archive_name': 'DEM_sample.csv'},
        )
        out_files.append('DEM_sample.zip')

    if 'Statistics' in out_reqs:
        demand_stats = describe(demand_sample)
        demand_stats = pd.concat([demand_stats, demand_units])
        demand_stats = convert_to_SimpleIndex(demand_stats, axis=1)
        demand_stats.to_csv(
            output_path / 'DEM_stats.csv',
            index_label=demand_stats.columns.name,
        )
        out_files.append('DEM_stats.csv')


def _asset_save(
    output_config: dict,
    assessment: DLCalculationAssessment,
    output_path: Path,
    out_files: list[str],
    *,
    aggregate_colocated: bool = False,
) -> None:
    """
    Save asset results to files based on the output config.

    Parameters
    ----------
    output_config : dict
        Configuration for output files.
    assessment : AssessmentBase
        The assessment object.
    output_path : Path
        Path to the output directory.
    out_files : list
        List of output file names.
    aggregate_colocated : bool, optional
        Whether to aggregate colocated components. Default is False.

    """
    output = assessment.asset.save_cmp_sample(save_units=True)
    assert isinstance(output, tuple)
    cmp_sample, cmp_units_series = output
    cmp_units = cmp_units_series.to_frame().T

    if aggregate_colocated:
        cmp_units = cmp_units.groupby(level=['cmp', 'loc', 'dir'], axis=1).first()  # type: ignore
        cmp_groupby_uid = cmp_sample.groupby(level=['cmp', 'loc', 'dir'], axis=1)  # type: ignore
        cmp_sample = cmp_groupby_uid.sum().mask(cmp_groupby_uid.count() == 0, np.nan)

    out_reqs = _parse_requested_output_file_names(output_config)

    if 'Sample' in out_reqs:
        cmp_sample_s = pd.concat([cmp_sample, cmp_units])

        cmp_sample_s = convert_to_SimpleIndex(cmp_sample_s, axis=1)
        cmp_sample_s.to_csv(
            output_path / 'CMP_sample.zip',
            index_label=cmp_sample_s.columns.name,
            compression={'method': 'zip', 'archive_name': 'CMP_sample.csv'},
        )
        out_files.append('CMP_sample.zip')

    if 'Statistics' in out_reqs:
        cmp_stats = describe(cmp_sample)
        cmp_stats = pd.concat([cmp_stats, cmp_units])

        cmp_stats = convert_to_SimpleIndex(cmp_stats, axis=1)
        cmp_stats.to_csv(
            output_path / 'CMP_stats.csv', index_label=cmp_stats.columns.name
        )
        out_files.append('CMP_stats.csv')


def _damage_save(
    output_config: dict,
    assessment: DLCalculationAssessment,
    output_path: Path,
    out_files: list[str],
    *,
    aggregate_colocated: bool = False,
    condense_ds: bool = False,
) -> None:
    """
    Save damage results to files based on the output config.

    Parameters
    ----------
    output_config : dict
        Configuration for output files.
    assessment : AssessmentBase
        The assessment object.
    output_path : Path
        Path to the output directory.
    out_files : list
        List of output file names.
    aggregate_colocated : bool, optional
        Whether to aggregate colocated components. Default is False.
    condense_ds : bool, optional
        Whether to condense damage states. Default is False.

    """
    output = assessment.damage.save_sample(save_units=True)
    assert isinstance(output, tuple)
    damage_sample, damage_units_series = output
    damage_units = damage_units_series.to_frame().T

    if aggregate_colocated:
        damage_units = damage_units.groupby(  # type: ignore
            level=['cmp', 'loc', 'dir', 'ds'], axis=1
        ).first()
        damage_groupby_uid = damage_sample.groupby(  # type: ignore
            level=['cmp', 'loc', 'dir', 'ds'], axis=1
        )
        damage_sample = damage_groupby_uid.sum().mask(
            damage_groupby_uid.count() == 0, np.nan
        )

    out_reqs = _parse_requested_output_file_names(output_config)

    if 'Sample' in out_reqs:
        damage_sample_s = pd.concat([damage_sample, damage_units])

        damage_sample_s = convert_to_SimpleIndex(damage_sample_s, axis=1)
        damage_sample_s.to_csv(
            output_path / 'DMG_sample.zip',
            index_label=damage_sample_s.columns.name,
            compression={
                'method': 'zip',
                'archive_name': 'DMG_sample.csv',
            },
        )
        out_files.append('DMG_sample.zip')

    if 'Statistics' in out_reqs:
        damage_stats = describe(damage_sample)
        damage_stats = pd.concat([damage_stats, damage_units])

        damage_stats = convert_to_SimpleIndex(damage_stats, axis=1)
        damage_stats.to_csv(
            output_path / 'DMG_stats.csv',
            index_label=damage_stats.columns.name,
        )
        out_files.append('DMG_stats.csv')

    if out_reqs.intersection({'GroupedSample', 'GroupedStatistics'}):
        damage_groupby = damage_sample.groupby(level=['cmp', 'loc', 'ds'], axis=1)  # type: ignore
        damage_units = damage_units.groupby(
            level=['cmp', 'loc', 'ds'], axis=1
        ).first()  # type: ignore

        grp_damage = damage_groupby.sum().mask(damage_groupby.count() == 0, np.nan)

        # if requested, condense DS output
        if condense_ds:
            # replace non-zero values with 1
            grp_damage = grp_damage.mask(
                grp_damage.astype(np.float64).to_numpy() > 0, 1
            )

            # get the corresponding DS for each column
            ds_list = grp_damage.columns.get_level_values('ds').astype(int)

            # replace ones with the corresponding DS in each cell
            grp_damage = grp_damage.mul(ds_list, axis=1)

            # aggregate across damage state indices
            damage_groupby_2 = grp_damage.groupby(level=['cmp', 'loc'], axis=1)

            # choose the max value
            # i.e., the governing DS for each comp-loc pair
            grp_damage = damage_groupby_2.max().mask(
                damage_groupby_2.count() == 0, np.nan
            )

            # aggregate units to the same format
            # assume identical units across locations for each comp
            damage_units = damage_units.groupby(level=['cmp', 'loc'], axis=1).first()  # type: ignore

        else:
            # otherwise, aggregate damage quantities for each comp
            damage_groupby_2 = grp_damage.groupby(level='cmp', axis=1)

            # preserve NaNs
            grp_damage = damage_groupby_2.sum().mask(
                damage_groupby_2.count() == 0, np.nan
            )

            # and aggregate units to the same format
            damage_units = damage_units.groupby(level='cmp', axis=1).first()  # type: ignore

        if 'GroupedSample' in out_reqs:
            grp_damage_s = pd.concat([grp_damage, damage_units])

            grp_damage_s = convert_to_SimpleIndex(grp_damage_s, axis=1)
            grp_damage_s.to_csv(
                output_path / 'DMG_grp.zip',
                index_label=grp_damage_s.columns.name,
                compression={
                    'method': 'zip',
                    'archive_name': 'DMG_grp.csv',
                },
            )
            out_files.append('DMG_grp.zip')

        if 'GroupedStatistics' in out_reqs:
            grp_stats = describe(grp_damage)
            grp_stats = pd.concat([grp_stats, damage_units])

            grp_stats = convert_to_SimpleIndex(grp_stats, axis=1)
            grp_stats.to_csv(
                output_path / 'DMG_grp_stats.csv',
                index_label=grp_stats.columns.name,
            )
            out_files.append('DMG_grp_stats.csv')


def _loss_save(  # noqa: C901
    output_config: dict,
    assessment: DLCalculationAssessment,
    output_path: Path,
    out_files: list[str],
    agg_repair: pd.DataFrame,
    *,
    aggregate_colocated: bool = False,
) -> None:
    """
    Save loss results to files based on the output config.

    Parameters
    ----------
    output_config : dict
        Configuration for output files.
    assessment : AssessmentBase
        The assessment object.
    output_path : Path
        Path to the output directory.
    out_files : list
        List of output file names.
    agg_repair : pd.DataFrame
        Aggregate repair data.
    aggregate_colocated : bool, optional
        Whether to aggregate colocated components. Default is False.

    """
    out = assessment.loss.save_sample(save_units=True)
    assert isinstance(out, tuple)
    repair_sample, repair_units_series = out
    repair_units = repair_units_series.to_frame().T

    if aggregate_colocated:
        if 'ds' in repair_units.columns.names:
            repair_units = repair_units.groupby(  # type: ignore
                level=['dv', 'loss', 'dmg', 'ds', 'loc', 'dir'], axis=1
            ).first()
            repair_groupby_uid = repair_sample.groupby(  # type: ignore
                level=['dv', 'loss', 'dmg', 'ds', 'loc', 'dir'], axis=1
            )
        else:
            repair_units = repair_units.groupby(  # type: ignore
                level=['dv', 'loss', 'dmg', 'loc', 'dir'], axis=1
            ).first()
            repair_groupby_uid = repair_sample.groupby(  # type: ignore
                level=['dv', 'loss', 'dmg', 'loc', 'dir'], axis=1
            )

        repair_sample = repair_groupby_uid.sum().mask(
            repair_groupby_uid.count() == 0, np.nan
        )

    out_reqs = _parse_requested_output_file_names(output_config)

    if 'Sample' in out_reqs:
        repair_sample_s = repair_sample.copy()
        repair_sample_s = pd.concat([repair_sample_s, repair_units])

        repair_sample_s = convert_to_SimpleIndex(repair_sample_s, axis=1)
        repair_sample_s.to_csv(
            output_path / 'DV_repair_sample.zip',
            index_label=repair_sample_s.columns.name,
            compression={
                'method': 'zip',
                'archive_name': 'DV_repair_sample.csv',
            },
        )
        out_files.append('DV_repair_sample.zip')

    if 'Statistics' in out_reqs:
        repair_stats = describe(repair_sample)
        repair_stats = pd.concat([repair_stats, repair_units])

        repair_stats = convert_to_SimpleIndex(repair_stats, axis=1)
        repair_stats.to_csv(
            output_path / 'DV_repair_stats.csv',
            index_label=repair_stats.columns.name,
        )
        out_files.append('DV_repair_stats.csv')

    if out_reqs.intersection({'GroupedSample', 'GroupedStatistics'}):
        repair_groupby = repair_sample.groupby(level=['dv', 'loss', 'dmg'], axis=1)  # type: ignore
        repair_units = repair_units.groupby(  # type: ignore
            level=['dv', 'loss', 'dmg'], axis=1
        ).first()
        grp_repair = repair_groupby.sum().mask(repair_groupby.count() == 0, np.nan)

        if 'GroupedSample' in out_reqs:
            grp_repair_s = pd.concat([grp_repair, repair_units])

            grp_repair_s = convert_to_SimpleIndex(grp_repair_s, axis=1)
            grp_repair_s.to_csv(
                output_path / 'DV_repair_grp.zip',
                index_label=grp_repair_s.columns.name,
                compression={
                    'method': 'zip',
                    'archive_name': 'DV_repair_grp.csv',
                },
            )
            out_files.append('DV_repair_grp.zip')

        if 'GroupedStatistics' in out_reqs:
            grp_stats = describe(grp_repair)
            grp_stats = pd.concat([grp_stats, repair_units])

            grp_stats = convert_to_SimpleIndex(grp_stats, axis=1)
            grp_stats.to_csv(
                output_path / 'DV_repair_grp_stats.csv',
                index_label=grp_stats.columns.name,
            )
            out_files.append('DV_repair_grp_stats.csv')

    if out_reqs.intersection({'AggregateSample', 'AggregateStatistics'}):
        if 'AggregateSample' in out_reqs:
            agg_repair_s = convert_to_SimpleIndex(agg_repair, axis=1)
            agg_repair_s.to_csv(
                output_path / 'DV_repair_agg.zip',
                index_label=agg_repair_s.columns.name,
                compression={
                    'method': 'zip',
                    'archive_name': 'DV_repair_agg.csv',
                },
            )
            out_files.append('DV_repair_agg.zip')

        if 'AggregateStatistics' in out_reqs:
            agg_stats = convert_to_SimpleIndex(describe(agg_repair), axis=1)
            agg_stats.to_csv(
                output_path / 'DV_repair_agg_stats.csv',
                index_label=agg_stats.columns.name,
            )
            out_files.append('DV_repair_agg_stats.csv')


def _remove_existing_files(output_path: Path, known_output_files: list[str]) -> None:
    """
    Remove known existing files from the specified output path.

    This function initializes the output folder by removing files that
    already exist in the `known_output_files` list.

    Parameters
    ----------
    output_path : Path
        The path to the output folder where files are located.
    known_output_files : list of str
        A list of filenames that are expected to exist and should be
        removed from the output folder.

    Raises
    ------
    OSError
        If an error occurs while attempting to remove a file, an
        OSError will be raised with the specific details of the
        failure.

    """
    # Initialize the output folder - i.e., remove existing output files from
    # there
    files = os.listdir(output_path)
    for filename in files:
        if filename in known_output_files:
            try:
                (output_path / filename).unlink()
            except OSError as exc:
                msg = (
                    f'Error occurred while removing '
                    f'`{output_path / filename}`: {exc}'
                )
                raise OSError(msg) from exc


def main() -> None:
    """Parse arguments and run the pelicun calculation."""
    args_list = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c',
        '--filenameDL',
        help='Path to the damage and loss (DL) configuration file.',
    )
    parser.add_argument(
        '-d',
        '--demandFile',
        default=None,
        help='Path to the file containing demand data.',
    )
    parser.add_argument(
        '-s',
        '--Realizations',
        default=None,
        help='Number of realizations to run in the probabilistic model.',
    )
    parser.add_argument(
        '--dirnameOutput',
        default=None,
        help='Directory where output files will be stored.',
    )
    parser.add_argument(
        '--detailed_results',
        default=True,
        type=str2bool,
        nargs='?',
        const=True,
        help='Generate detailed results (True/False). Defaults to True.',
    )
    parser.add_argument(
        '--coupled_EDP',
        default=False,
        type=str2bool,
        nargs='?',
        const=False,
        help=(
            'Consider coupled Engineering Demand Parameters (EDPs) '
            'in calculations (True/False). Defaults to False.'
        ),
    )
    parser.add_argument(
        '--log_file',
        default=True,
        type=str2bool,
        nargs='?',
        const=True,
        help='Generate a log file (True/False). Defaults to True.',
    )
    parser.add_argument(
        '--auto_script',
        default=None,
        help='Optional path to a config auto-generation script.',
    )
    parser.add_argument(
        '--custom_model_dir',
        default=None,
        help='Directory containing custom model data.',
    )
    parser.add_argument(
        '--output_format',
        default=None,
        help='Desired output format for the results.',
    )
    # TODO(JVM): fix color warnings
    # parser.add_argument(
    #     '--color_warnings',
    #     default=False,
    #     type=str2bool,
    #     nargs='?',
    #     const=False,
    #     help=(
    #         'Enable colored warnings in the console '
    #         'output (True/False). Defaults to False.'
    #     ),
    # )
    parser.add_argument(
        '--ground_failure',
        default=False,
        type=str2bool,
        nargs='?',
        const=False,
        help='Currently not used. Soon to be deprecated.',
    )
    parser.add_argument(
        '--regional',
        default=False,
        type=str2bool,
        nargs='?',
        const=False,
        help='Currently not used. Soon to be deprecated.',
    )
    parser.add_argument('--resource_dir', default=None)

    if not args_list:
        parser.print_help()
        return

    args = parser.parse_args(args_list)

    log_msg('Initializing pelicun calculation.')

    run_pelicun(
        config_path=args.filenameDL,
        demand_file=args.demandFile,
        output_path=args.dirnameOutput,
        realizations=args.Realizations,
        auto_script_path=args.auto_script,
        custom_model_dir=args.custom_model_dir,
        output_format=args.output_format,
        detailed_results=args.detailed_results,
        coupled_edp=args.coupled_EDP,
    )

    log_msg('pelicun calculation completed.')


if __name__ == '__main__':
    main()



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the demand model of pelicun."""

from __future__ import annotations

import tempfile
import warnings
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

from pelicun.base import ensure_value
from pelicun.model.demand_model import (
    DemandModel,
    _assemble_required_demand_data,
    _get_required_demand_type,
)
from pelicun.tests.basic.test_model import TestModelModule

if TYPE_CHECKING:
    from pelicun.assessment import Assessment


class TestDemandModel(TestModelModule):  # noqa: PLR0904
    @pytest.fixture
    def demand_model(self, assessment_instance: Assessment) -> DemandModel:
        return deepcopy(assessment_instance.demand)

    @pytest.fixture
    def demand_model_with_sample(
        self, assessment_instance: Assessment
    ) -> DemandModel:
        mdl = assessment_instance.demand
        mdl.load_sample(
            'pelicun/tests/basic/data/model/'
            'test_DemandModel/load_sample/demand_sample_A.csv'
        )
        model_copy = deepcopy(mdl)
        assert isinstance(model_copy, DemandModel)
        return model_copy

    @pytest.fixture
    def calibrated_demand_model(
        self, demand_model_with_sample: DemandModel
    ) -> DemandModel:
        config = {
            'ALL': {
                'DistributionFamily': 'normal',
                'AddUncertainty': 0.00,
            },
            'PID': {
                'DistributionFamily': 'lognormal',
                'TruncateUpper': '0.06',
            },
            'SA': {
                'DistributionFamily': 'empirical',
            },
        }
        demand_model_with_sample.calibrate_model(config)
        model_copy = deepcopy(demand_model_with_sample)
        assert isinstance(model_copy, DemandModel)
        return model_copy

    @pytest.fixture
    def demand_model_with_sample_b(
        self, assessment_instance: Assessment
    ) -> DemandModel:
        mdl = assessment_instance.demand
        mdl.load_sample(
            'pelicun/tests/basic/data/model/'
            'test_DemandModel/load_sample/demand_sample_B.csv'
        )
        model_copy = deepcopy(mdl)
        assert isinstance(model_copy, DemandModel)
        return model_copy

    @pytest.fixture
    def demand_model_with_sample_c(
        self, assessment_instance: Assessment
    ) -> DemandModel:
        mdl = assessment_instance.demand
        mdl.load_sample(
            'pelicun/tests/basic/data/model/'
            'test_DemandModel/load_sample/demand_sample_C.csv'
        )
        model_copy = deepcopy(mdl)
        assert isinstance(model_copy, DemandModel)
        return model_copy

    @pytest.fixture
    def demand_model_with_sample_d(
        self, assessment_instance: Assessment
    ) -> DemandModel:
        mdl = assessment_instance.demand
        mdl.load_sample(
            'pelicun/tests/basic/data/model/'
            'test_DemandModel/load_sample/demand_sample_D.csv'
        )
        model_copy = deepcopy(mdl)
        assert isinstance(model_copy, DemandModel)
        return model_copy

    def test_init(self, demand_model: DemandModel) -> None:
        assert demand_model.log

        assert demand_model.marginal_params is None
        assert demand_model.correlation is None
        assert demand_model.empirical_data is None
        assert demand_model.user_units is None
        assert demand_model._RVs is None
        assert demand_model.sample is None

    def test_save_sample(self, demand_model_with_sample: DemandModel) -> None:
        # instantiate a temporary directory in memory
        temp_dir = tempfile.mkdtemp()
        # save the sample there
        demand_model_with_sample.save_sample(f'{temp_dir}/temp.csv')
        with Path(f'{temp_dir}/temp.csv').open(encoding='utf-8') as f:
            contents = f.read()
        assert contents == (
            ',PFA-0-1,PFA-1-1,PID-1-1,SA_0.23-0-1\n'
            'Units,inps2,inps2,rad,inps2\n'
            '0,158.62478,397.04389,0.02672,342.149\n'
        )
        res = demand_model_with_sample.save_sample()
        assert isinstance(res, pd.DataFrame)
        assert res.to_dict() == {
            ('PFA', '0', '1'): {0: 158.62478},
            ('PFA', '1', '1'): {0: 397.04389},
            ('PID', '1', '1'): {0: 0.02672},
            ('SA_0.23', '0', '1'): {0: 342.149},
        }

    def test_load_sample(
        self,
        demand_model_with_sample: DemandModel,
        demand_model_with_sample_b: DemandModel,
    ) -> None:
        # retrieve the loaded sample and units
        obtained_sample = ensure_value(demand_model_with_sample.sample)
        obtained_units = ensure_value(demand_model_with_sample.user_units)

        obtained_sample_2 = ensure_value(demand_model_with_sample_b.sample)
        obtained_units_2 = ensure_value(demand_model_with_sample_b.user_units)

        # demand_sample_A.csv and demand_sample_b.csv only differ in the
        # headers, where the first includes a tag for the hazard
        # level. Therefore, the two files are expected to result to the
        # same `obtained_sample`

        pd.testing.assert_frame_equal(
            obtained_sample,
            obtained_sample_2,
            check_index_type=False,
            check_column_type=False,
        )
        pd.testing.assert_series_equal(
            obtained_units,
            obtained_units_2,
            check_index_type=False,
        )

        # compare against the expected values for the sample
        expected_sample = pd.DataFrame(
            {
                ('PFA', '0', '1'): [4.029069],
                ('PFA', '1', '1'): [10.084915],
                ('PID', '1', '1'): [0.02672],
                ('SA_0.23', '0', '1'): [8.690585],
            },
            index=[0],
        ).rename_axis(columns=['type', 'loc', 'dir'])
        pd.testing.assert_frame_equal(
            expected_sample,
            obtained_sample,
            check_index_type=False,
            check_column_type=False,
        )

        # compare against the expected values for the units
        expected_units = pd.Series(
            ('inps2', 'inps2', 'rad', 'inps2'),
            index=pd.MultiIndex.from_tuples(
                (
                    ('PFA', '0', '1'),
                    ('PFA', '1', '1'),
                    ('PID', '1', '1'),
                    ('SA_0.23', '0', '1'),
                ),
                names=['type', 'loc', 'dir'],
            ),
            name='Units',
        )
        pd.testing.assert_series_equal(
            expected_units,
            obtained_units,
            check_index_type=False,
        )

    def test_estimate_RID(self, demand_model_with_sample: DemandModel) -> None:
        demands = ensure_value(demand_model_with_sample.sample)['PID']
        params = {'yield_drift': 0.01}
        res = demand_model_with_sample.estimate_RID(demands, params)
        assert list(res.columns) == [('RID', '1', '1')]
        with pytest.raises(ValueError, match='Invalid method: `xyz`'):
            demand_model_with_sample.estimate_RID(demands, params, method='xyz')

    def test_expand_sample_float(
        self, demand_model_with_sample: DemandModel
    ) -> None:
        sample_before = ensure_value(demand_model_with_sample.sample).copy()
        demand_model_with_sample.expand_sample('test_lab', 1.00, 'unitless')
        sample_after = ensure_value(demand_model_with_sample.sample).copy()
        pd.testing.assert_frame_equal(
            sample_before, sample_after.drop('test_lab', axis=1)
        )
        assert sample_after.loc[0, ('test_lab', '0', '1')] == 1.0

    def test_expand_sample_numpy(
        self, demand_model_with_sample: DemandModel
    ) -> None:
        sample_before = ensure_value(demand_model_with_sample.sample).copy()
        demand_model_with_sample.expand_sample('test_lab', 1.00, 'unitless')
        sample_after = ensure_value(demand_model_with_sample.sample).copy()
        pd.testing.assert_frame_equal(
            sample_before, sample_after.drop('test_lab', axis=1)
        )
        assert sample_after.loc[0, ('test_lab', '0', '1')] == 1.0

    def test_expand_sample_error_no_sample(self, demand_model: DemandModel) -> None:
        with pytest.raises(
            ValueError, match='Demand model does not have a sample yet.'
        ):
            demand_model.expand_sample('test_lab', np.array((1.00,)), 'unitless')

    def test_expand_sample_error_wrong_shape(
        self, demand_model_with_sample: DemandModel
    ) -> None:
        with pytest.raises(ValueError, match='Incompatible array length.'):
            demand_model_with_sample.expand_sample(
                'test_lab', np.array((1.00, 1.00)), 'unitless'
            )

    def test_calibrate_model(
        self,
        calibrated_demand_model: DemandModel,
    ) -> None:
        assert ensure_value(calibrated_demand_model.marginal_params)[
            'Family'
        ].to_list() == [
            'normal',
            'normal',
            'lognormal',
            'empirical',
        ]
        assert (
            ensure_value(calibrated_demand_model.marginal_params).loc[
                ('PID', '1', '1'), 'TruncateUpper'
            ]
            == 0.06
        )

    def test_calibrate_model_censoring(
        self,
        demand_model_with_sample_c: DemandModel,
    ) -> None:
        # with a config featuring censoring the RIDs
        config = {
            'ALL': {
                'DistributionFamily': 'normal',
                'AddUncertainty': 0.00,
            },
            'PID': {
                'DistributionFamily': 'lognormal',
                'CensorUpper': '0.05',
            },
        }
        demand_model_with_sample_c.calibrate_model(config)

    def test_calibrate_model_truncation(
        self,
        demand_model_with_sample_c: DemandModel,
    ) -> None:
        # with a config that specifies a truncation limit smaller than
        # the samples
        config = {
            'ALL': {
                'DistributionFamily': 'normal',
                'AddUncertainty': 0.00,
            },
            'PID': {
                'DistributionFamily': 'lognormal',
                'TruncateUpper': '0.04',
            },
        }
        demand_model_with_sample_c.calibrate_model(config)

    def test_save_load_model_with_empirical(
        self, calibrated_demand_model: DemandModel, assessment_instance: Assessment
    ) -> None:
        # a model that has empirical marginal parameters
        temp_dir = tempfile.mkdtemp()
        calibrated_demand_model.save_model(f'{temp_dir}/temp')
        assert Path(f'{temp_dir}/temp_marginals.csv').exists()
        assert Path(f'{temp_dir}/temp_empirical.csv').exists()
        assert Path(f'{temp_dir}/temp_correlation.csv').exists()

        # Load model to a different DemandModel instance to verify
        new_demand_model = assessment_instance.demand
        new_demand_model.load_model(f'{temp_dir}/temp')
        pd.testing.assert_frame_equal(
            ensure_value(calibrated_demand_model.marginal_params),
            ensure_value(new_demand_model.marginal_params),
            atol=1e-4,
            check_index_type=False,
            check_column_type=False,
        )
        pd.testing.assert_frame_equal(
            ensure_value(calibrated_demand_model.correlation),
            ensure_value(new_demand_model.correlation),
            atol=1e-4,
            check_index_type=False,
            check_column_type=False,
        )
        pd.testing.assert_frame_equal(
            ensure_value(calibrated_demand_model.empirical_data),
            ensure_value(new_demand_model.empirical_data),
            atol=1e-4,
            check_index_type=False,
            check_column_type=False,
        )

    def test_save_load_model_without_empirical(
        self,
        demand_model_with_sample_c: DemandModel,
        assessment_instance: Assessment,
    ) -> None:
        # a model that does not have empirical marginal parameters
        temp_dir = tempfile.mkdtemp()
        config = {
            'ALL': {
                'DistributionFamily': 'normal',
                'AddUncertainty': 0.00,
            },
            'PID': {
                'DistributionFamily': 'lognormal',
                'TruncateUpper': '0.04',
            },
        }
        demand_model_with_sample_c.calibrate_model(config)
        demand_model_with_sample_c.save_model(f'{temp_dir}/temp')
        assert Path(f'{temp_dir}/temp_marginals.csv').exists()
        assert Path(f'{temp_dir}/temp_correlation.csv').exists()

        # Load model to a different DemandModel instance to verify
        new_demand_model = assessment_instance.demand
        new_demand_model.load_model(f'{temp_dir}/temp')
        pd.testing.assert_frame_equal(
            ensure_value(demand_model_with_sample_c.marginal_params),
            ensure_value(new_demand_model.marginal_params),
        )
        pd.testing.assert_frame_equal(
            ensure_value(demand_model_with_sample_c.correlation),
            ensure_value(new_demand_model.correlation),
        )
        assert demand_model_with_sample_c.empirical_data is None
        assert new_demand_model.empirical_data is None

    def test_generate_sample_exceptions(self, demand_model: DemandModel) -> None:
        # generating a sample from a non calibrated model should fail
        with pytest.raises(
            ValueError, match='Model parameters have not been specified'
        ):
            demand_model.generate_sample(
                {'SampleSize': 3, 'PreserveRawOrder': False}
            )

    def test_generate_sample(self, calibrated_demand_model: DemandModel) -> None:
        calibrated_demand_model.generate_sample(
            {'SampleSize': 3, 'PreserveRawOrder': False}
        )

        # get the generated demand sample
        res = calibrated_demand_model.save_sample(save_units=True)
        assert isinstance(res, tuple)

        obtained_sample, obtained_units = res

        # compare against the expected values for the sample
        expected_sample = pd.DataFrame(
            {
                ('PFA', '0', '1'): [158.624160, 158.624160, 158.624160],
                ('PFA', '1', '1'): [397.042985, 397.042985, 397.042985],
                ('PID', '1', '1'): [0.02672, 0.02672, 0.02672],
                ('SA_0.23', '0', '1'): [342.148783, 342.148783, 342.148783],
            },
            index=pd.Index([0, 1, 2], dtype='object'),
        ).rename_axis(columns=['type', 'loc', 'dir'])
        pd.testing.assert_frame_equal(
            expected_sample,
            obtained_sample,
            check_exact=False,
            atol=1e-4,
            check_index_type=False,
            check_column_type=False,
        )

        # compare against the expected values for the units
        expected_units = pd.Series(
            ('inps2', 'inps2', 'rad', 'inps2'),
            index=pd.MultiIndex.from_tuples(
                (
                    ('PFA', '0', '1'),
                    ('PFA', '1', '1'),
                    ('PID', '1', '1'),
                    ('SA_0.23', '0', '1'),
                ),
                names=('type', 'loc', 'dir'),
            ),
            name='Units',
        )
        pd.testing.assert_series_equal(
            expected_units,
            obtained_units,
            check_index_type=False,
        )

    def test_generate_sample_with_demand_cloning(
        self, assessment_instance: Assessment
    ) -> None:
        # # used for debugging
        # assessment_instance = assessment.Assessment()

        demand_model = assessment_instance.demand

        mdl = assessment_instance.demand
        # contains PGV-0-1, PGV-1-1, PGV-2-1, and PGA-0-1
        # PGA-0-1 is not cloned.
        mdl.load_sample(
            'pelicun/tests/basic/data/model/'
            'test_DemandModel/generate_sample_with_demand_cloning/sample.csv'
        )
        demand_model.calibrate_model(
            {
                'ALL': {
                    'DistributionFamily': 'lognormal',
                },
            }
        )
        with warnings.catch_warnings(record=True) as w:
            demand_model.generate_sample(
                {
                    'SampleSize': 1000,
                    'DemandCloning': {
                        'PGV-0-1': ['PGV-0-1', 'PGV-0-2', 'PGV-0-3'],
                        'PGV-1-1': ['PGV-1-1', 'PGV-1-2', 'PGV-1-3'],
                        'PGV-2-1': ['PGV-2-1', 'PGV-2-2', 'PGV-2-3'],
                        'not_present': ['X-0-0', 'Y-0-0', 'Z-0-0'],
                    },
                }
            )
        assert len(w) == 1
        assert (
            'The demand cloning configuration lists columns '
            "that are not present in the original demand sample's "
            "columns: ['not_present']."
        ) in str(w[0].message)
        # we'll just get a warning for the `not_present` entry
        assert ensure_value(demand_model.sample).columns.to_list() == [
            ('PGA', '0', '1'),
            ('PGV', '0', '1'),
            ('PGV', '0', '2'),
            ('PGV', '0', '3'),
            ('PGV', '1', '1'),
            ('PGV', '1', '2'),
            ('PGV', '1', '3'),
            ('PGV', '2', '1'),
            ('PGV', '2', '2'),
            ('PGV', '2', '3'),
        ]
        assert np.array_equal(
            demand_model.sample['PGV', '0', '1'].values,  # type: ignore
            demand_model.sample['PGV', '0', '3'].values,  # type: ignore
        )
        # exceptions
        # Duplicate entries in demand cloning configuration
        with pytest.raises(
            ValueError, match='Duplicate entries in demand cloning configuration.'
        ):
            demand_model.generate_sample(
                {
                    'SampleSize': 1000,
                    'DemandCloning': {
                        'PGV-0-1': ['PGV-0-1', 'PGV-0-2', 'PGV-0-3'],
                        'PGV-1-1': ['PGV-0-1', 'PGV-1-2', 'PGV-1-3'],
                        'PGV-2-1': ['PGV-0-1', 'PGV-2-2', 'PGV-2-3'],
                    },
                }
            )

    def test__get_required_demand_type(
        self, assessment_instance: Assessment
    ) -> None:
        # Simple case: single demand
        damage_model = assessment_instance.damage
        cmp_set = {'testing.component'}
        damage_model.load_model_parameters(
            [
                'pelicun/tests/basic/data/model/test_DemandModel/'
                '_get_required_demand_type/damage_db_testing_single.csv'
            ],
            cmp_set,
        )
        pgb = pd.DataFrame(
            {('testing.component', '1', '1', '1'): [1]}, index=['Blocks']
        ).T.rename_axis(index=['cmp', 'loc', 'dir', 'uid'])
        demand_offset = {'PFA': 0}
        required = _get_required_demand_type(
            ensure_value(damage_model.ds_model.damage_params), pgb, demand_offset
        )
        expected = defaultdict(
            list,
            {(('PID-1-1',), None): [('testing.component', '1', '1', '1')]},
        )
        assert required == expected

        # Utility demand case: two demands are required
        damage_model = assessment_instance.damage
        cmp_set = {'testing.component2'}
        damage_model.load_model_parameters(
            [
                'pelicun/tests/basic/data/model/test_DemandModel/'
                '_get_required_demand_type/damage_db_testing_utility.csv'
            ],
            cmp_set,
        )
        pgb = pd.DataFrame(
            {('testing.component2', '1', '1', '1'): [1]}, index=['Blocks']
        ).T.rename_axis(index=['cmp', 'loc', 'dir', 'uid'])
        demand_offset = {'PFA': 0}
        required = _get_required_demand_type(
            ensure_value(damage_model.ds_model.damage_params), pgb, demand_offset
        )
        expected = defaultdict(
            list,
            {
                (('PID-1-1', 'PFA-1-1'), 'sqrt(X1^2+X2^2)'): [  # type: ignore
                    ('testing.component2', '1', '1', '1')
                ]
            },
        )
        assert required == expected

    def test__assemble_required_demand_data(
        self, assessment_instance: Assessment
    ) -> None:
        # Utility demand case: two demands are required
        damage_model = assessment_instance.damage
        cmp_set = {'testing.component'}
        damage_model.load_model_parameters(
            [
                'pelicun/tests/basic/data/model/test_DemandModel/'
                '_get_required_demand_type/damage_db_testing_single.csv'
            ],
            cmp_set,
        )
        required_edps = defaultdict(
            list,
            {
                (('PID-1-1', 'PFA-1-1'), 'sqrt(X1^2+X2^2)'): [
                    ('testing.component', '1', '1', '1')
                ]
            },
        )
        nondirectional_multipliers = {'ALL': 1.00}
        demand_sample = pd.DataFrame(
            {
                ('PID', '1', '1'): np.full(5, 3.00),
                ('PFA', '1', '1'): np.full(5, 4.00),
            }
        )
        demand_data = _assemble_required_demand_data(
            required_edps,  # type: ignore
            nondirectional_multipliers,
            demand_sample,
        )
        expected = {
            (('PID-1-1', 'PFA-1-1'), 'sqrt(X1^2+X2^2)'): np.array(
                [5.0, 5.0, 5.0, 5.0, 5.0]
            )
        }
        assert demand_data.keys() == expected.keys()
        for key in demand_data:
            assert np.all(demand_data[key] == expected[key])



# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

"""Code inspection methods/functions."""

from __future__ import annotations  # noqa: I001
from pathlib import Path

import ast


def visit_FunctionDef(
    node: ast.FunctionDef,
    filename: str,
    search_string: str,
    functions_with_string: list[str],
) -> None:
    """
    Visit a function definition node and check if it contains the
    search string.

    Parameters
    ----------
    node: ast.FunctionDef
        The AST node representing the function definition.
    filename: str
        The path to the Python file to be searched.
    search_string: str
        The string to search for within the function bodies.
    functions_with_string: list[str]
        The list to append function names that contain the search
        string.

    """
    with Path(filename).open(encoding='utf-8') as f:
        contents = f.read()

    function_code = ast.get_source_segment(contents, node)
    assert function_code is not None

    if search_string in function_code:
        functions_with_string.append(node.name)


def find_functions_with_string(filename: str, search_string: str) -> list[str]:
    """
    Finds functions in a Python file that contain a specific string in
    their body.

    Parameters
    ----------
    filename: str
        The path to the Python file to be searched.
    search_string: str
        The string to search for within the function bodies.

    Returns
    -------
    list[str]
        A list of function names that contain the search string in
        their bodies.

    """
    with Path(filename).open(encoding='utf-8') as file:
        contents = file.read()
    tree = ast.parse(contents, filename=filename)

    functions_with_string: list[str] = []

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            visit_FunctionDef(node, filename, search_string, functions_with_string)

    return functions_with_string



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis


"""Constants, basic classes, and methods for pelicun."""

from __future__ import annotations

import argparse
import json
import pprint
import sys
import traceback
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypeVar, overload

import colorama
import numpy as np
import pandas as pd
from colorama import Fore, Style
from scipy.interpolate import interp1d  # type: ignore

from pelicun.pelicun_warnings import PelicunWarning

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType

    from pelicun.assessment import AssessmentBase


colorama.init()
# set printing options
pp = pprint.PrettyPrinter(indent=2, width=80 - 24)

pd.options.display.max_rows = 20
pd.options.display.max_columns = None  # type: ignore
pd.options.display.expand_frame_repr = True
pd.options.display.width = 300

idx = pd.IndexSlice


T = TypeVar('T')


class Options:
    """
    Analysis options and logging configuration.

    Attributes
    ----------
    sampling_method: str
        Sampling method to use. Specified in the user's configuration
        dictionary, otherwise left as provided in the default configuration
        file (see settings/default_config.json in the pelicun source
        code). Can be any of ['LHS', 'LHS_midpoint',
        'MonteCarlo']. The default is 'LHS'.
    units_file: str
        Location of a user-specified units file, which should contain
        the names of supported units and their conversion factors (the
        value some quantity of a given unit needs to be multiplied to
        be expressed in the base units). Value specified in the user
        configuration dictionary. Pelicun comes with a set of default
        units which are always loaded (see
        `settings/default_units.json` in the pelicun source
        code). Units specified in the units_file overwrite the default
        units.
    demand_offset: dict
        Demand offsets are used in the process of mapping a component
        location to its associated EDP. This allows components that
        are sensitive to EDPs of different levels to be specified as
        present at the same location (e.g. think of desktop computer
        and suspended ceiling, both at the same story). Each
        component's offset value is specified in the component
        fragility database. This setting applies a supplemental global
        offset to specific EDP types. The value is specified in the
        user's configuration dictionary, otherwise left as provided in
        the default configuration file (see
        settings/default_config.json in the pelicun source code).
    nondir_multi_dict: dict
        Nondirectional components are sensitive to demands coming in
        any direction. Results are typically available in two
        orthogonal directions. FEMA P-58 suggests using the formula
        `max(dir_1, dir_2) * 1.2` to estimate the demand for such
        components. This parameter allows modifying the 1.2 multiplier
        with a user-specified value. The change can be applied to
        "ALL" EDPs, or for specific EDPs, such as "PFA", "PFV",
        etc. The value is specified in the user's configuration
        dictionary, otherwise left as provided in the default
        configuration file (see settings/default_config.json in the
        pelicun source code).
    rho_cost_time: float
        Specifies the correlation between the repair cost and repair
        time consequences. The value is specified in the user's
        configuration dictionary, otherwise left as provided in the
        default configuration file (see
        "RepairCostAndTimeCorrelation") (see
        settings/default_config.json in the pelicun source code).
    eco_scale: dict
        Controls how the effects of economies of scale are handled in
        the damaged component quantity aggregation for loss measure
        estimation. The dictionary is specified in the user's
        configuration dictionary, otherwise left as provided in the
        default configuration file (see settings/default_config.json
        in the pelicun source code).
    log: Logger
        Logger object. Configuration parameters coming from the user's
        configuration dictionary or the default configuration file
        control logging behavior. See Logger class.

    """

    __slots__ = [
        '_asmnt',
        '_rng',
        '_seed',
        'defaults',
        'demand_offset',
        'eco_scale',
        'eco_scale',
        'error_setup',
        'error_setup',
        'list_all_ds',
        'log',
        'log',
        'nondir_multi_dict',
        'rho_cost_time',
        'sampling_method',
        'units_file',
    ]

    def __init__(
        self,
        user_config_options: dict[str, Any] | None,
        assessment: AssessmentBase | None = None,
    ) -> None:
        """
        Initialize an Options object.

        Parameters
        ----------
        user_config_options: dict, Optional
            User-specified configuration dictionary. Any provided
            user_config_options override the defaults.
        assessment: AssessmentBase, Optional
            Assessment object that will be using this Options
            object. If it is not intended to use this Options object
            for an Assessment (e.g. defining an Options object for UQ
            use), this value should be None.

        """
        self._asmnt = assessment

        self.defaults: dict[str, Any] | None = None
        self.sampling_method: str | None = None
        self.list_all_ds: bool | None = None

        merged_config_options = merge_default_config(user_config_options)

        self.seed = merged_config_options['Seed']
        self.sampling_method = merged_config_options['Sampling']['SamplingMethod']
        self.list_all_ds = merged_config_options['ListAllDamageStates']

        self.units_file = merged_config_options['UnitsFile']

        self.demand_offset = merged_config_options['DemandOffset']
        self.nondir_multi_dict = merged_config_options['NonDirectionalMultipliers']
        self.rho_cost_time = merged_config_options['RepairCostAndTimeCorrelation']
        self.eco_scale = merged_config_options['EconomiesOfScale']

        self.error_setup = merged_config_options['ErrorSetup']

        # instantiate a Logger object with the finalized configuration
        self.log = Logger(
            merged_config_options['LogFile'],
            verbose=merged_config_options['Verbose'],
            log_show_ms=merged_config_options['LogShowMS'],
            print_log=merged_config_options['PrintLog'],
        )

    @property
    def seed(self) -> float | None:
        """
        Seed property.

        Returns
        -------
        float
            Seed value

        """
        return self._seed

    @seed.setter
    def seed(self, value: float) -> None:
        """Seed property setter."""
        self._seed = value
        self._rng = np.random.default_rng(self._seed)  # type: ignore

    @property
    def rng(self) -> np.random.Generator:
        """
        rng property.

        Returns
        -------
        Generator
            Random generator

        """
        return self._rng


# Define a module-level LoggerRegistry
class LoggerRegistry:
    """Registry to manage all logger instances."""

    _loggers: ClassVar[list[Logger]] = []

    # The @classmethod decorator allows this method to be called on
    # the class itself, rather than on instances. It interacts with
    # class-level data (like _loggers), enabling a single registry for
    # all Logger instances without needing an object of LoggerRegistry
    # itself.
    @classmethod
    def register(cls, logger: Logger) -> None:
        """Register a logger instance."""
        cls._loggers.append(logger)

    @classmethod
    def log_exception(
        cls,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Log exceptions to all registered loggers."""
        message = (
            f"Unhandled exception occurred:"
            f"\n"
            f"{''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))}"
        )
        for logger in cls._loggers:
            logger.msg(message)

        # Also call the default excepthook to print the exception to
        # the console as is done by default.
        sys.__excepthook__(exc_type, exc_value, exc_traceback)


# Update sys.excepthook to log exceptions in all loggers
# https://docs.python.org/3/library/sys.html#sys.excepthook
sys.excepthook = LoggerRegistry.log_exception


class Logger:
    """Generate log files documenting execution events."""

    __slots__ = [
        'emitted',
        'log_div',
        'log_file',
        'log_show_ms',
        'log_time_format',
        'print_log',
        'spaces',
        'verbose',
        'warning_file',
        'warning_stack',
    ]

    def __init__(
        self,
        log_file: str | None,
        *,
        verbose: bool,
        log_show_ms: bool,
        print_log: bool,
    ) -> None:
        """
        Initialize a Logger object.

        Parameters
        ----------
        verbose: bool
            If True, the pelicun echoes more information throughout the
            assessment.  This can be useful for debugging purposes. The
            value is specified in the user's configuration dictionary,
            otherwise left as provided in the default configuration file
            (see settings/default_config.json in the pelicun source code).
        log_show_ms: bool
            If True, the timestamps in the log file are in microsecond
            precision. The value is specified in the user's configuration
            dictionary, otherwise left as provided in the default
            configuration file (see settings/default_config.json in the
            pelicun source code).
        log_file: str, optional
            If a value is provided, the log is written to that file. The
            value is specified in the user's configuration dictionary,
            otherwise left as provided in the default configuration file
            (see settings/default_config.json in the pelicun source code).
        print_log: bool
            If True, the log is also printed to standard output. The
            value is specified in the user's configuration dictionary,
            otherwise left as provided in the default configuration file
            (see settings/default_config.json in the pelicun source code).

        """
        self.verbose = verbose
        self.log_show_ms = bool(log_show_ms)

        if log_file is None:
            self.log_file = None
            self.warning_file = None
        else:
            path = Path(log_file)
            self.log_file = str(path.resolve())
            name, extension = split_file_name(self.log_file)
            self.warning_file = (
                path.parent / (name + '_warnings' + extension)
            ).resolve()
            with Path(self.log_file).open('w', encoding='utf-8') as f:
                f.write('')
            with Path(self.warning_file).open('w', encoding='utf-8') as f:
                f.write('')

        self.print_log = str2bool(print_log)
        self.warning_stack: list[str] = []
        self.emitted: set[str] = set()
        self.reset_log_strings()
        control_warnings()

        # Register the logger to the LoggerRegistry in order to
        # capture raised exceptions.
        LoggerRegistry.register(self)

    def reset_log_strings(self) -> None:
        """Populate the string-related attributes of the logger."""
        if self.log_show_ms:
            self.log_time_format = '%H:%M:%S:%f'
            # the length of the time string in the log file
            self.spaces = ' ' * 16
            # to have a total length of 80 with the time added
            self.log_div = '-' * (80 - 17)
        else:
            self.log_time_format = '%H:%M:%S'
            self.spaces = ' ' * 9
            self.log_div = '-' * (80 - 10)

    def msg(
        self,
        msg: str = '',
        *,
        prepend_timestamp: bool = True,
        prepend_blank_space: bool = True,
    ) -> None:
        """
        Write a message in the log file with the current time as prefix.

        The time is in ISO-8601 format, e.g. 2018-06-16T20:24:04Z

        Parameters
        ----------
        msg: string
            Message to print.
        prepend_timestamp: bool
            Controls whether a timestamp is placed before the message.
        prepend_blank_space: bool
            Controls whether blank space is placed before the message.

        """
        msg_lines = msg.split('\n')

        for msg_i, msg_line in enumerate(msg_lines):
            if prepend_timestamp and (msg_i == 0):
                formatted_msg = (
                    f'{datetime.now().strftime(self.log_time_format)} {msg_line}'  # noqa: DTZ005
                )
            elif prepend_timestamp or prepend_blank_space:
                formatted_msg = self.spaces + msg_line
            else:
                formatted_msg = msg_line

            if self.print_log:
                print(formatted_msg)  # noqa: T201

            if self.log_file is not None:
                with Path(self.log_file).open('a', encoding='utf-8') as f:
                    f.write('\n' + formatted_msg)

    def add_warning(self, msg: str) -> None:
        """
        Add a warning to the warning stack.

        Notes
        -----
        Warnings are only emitted when `emit_warnings` is called.

        Parameters
        ----------
        msg: str
            The warning message.

        """
        msg_lines = msg.split('\n')
        formatted_msg = '\n'
        for msg_line in msg_lines:
            formatted_msg += (
                self.spaces + Fore.RED + msg_line + Style.RESET_ALL + '\n'
            )
        if formatted_msg not in self.warning_stack:
            self.warning_stack.append(formatted_msg)

    def emit_warnings(self) -> None:
        """Issues all warnings and clears the warning stack."""
        for message in self.warning_stack:
            if message not in self.emitted:
                warnings.warn(message, PelicunWarning, stacklevel=3)
                if self.warning_file is not None:
                    with Path(self.warning_file).open('a', encoding='utf-8') as f:
                        f.write(
                            message.replace(Fore.RED, '')
                            .replace(Style.RESET_ALL, '')
                            .replace(self.spaces, '')
                        )

        self.emitted = self.emitted.union(set(self.warning_stack))
        self.warning_stack = []

    def warning(self, msg: str) -> None:
        """
        Add an emit a warning immediately.

        Parameters
        ----------
        msg: str
            Warning message

        """
        self.add_warning(msg)
        self.emit_warnings()

    def div(self, *, prepend_timestamp: bool = False) -> None:
        """Add a divider line in the log file."""
        msg = self.log_div if prepend_timestamp else '-' * 80
        self.msg(msg, prepend_timestamp=prepend_timestamp)

    def print_system_info(self) -> None:
        """Write system information in the log."""
        self.msg(
            'System Information:', prepend_timestamp=False, prepend_blank_space=False
        )
        start = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')  # noqa: DTZ005
        self.msg(
            f'local time zone: {datetime.now(timezone.utc).astimezone().tzinfo}\n'
            f'start time: {start}\n'
            f'python: {sys.version}\n'
            f'numpy: {np.__version__}\n'
            f'pandas: {pd.__version__}\n',
            prepend_timestamp=False,
        )


# get the absolute path of the pelicun directory
pelicun_path = Path(__file__).resolve().parent


def split_file_name(file_path: str) -> tuple[str, str]:
    """
    Separate a file name from the extension.

    Separates a file name from the extension accounting for the case
    where the file name itself contains periods.

    Parameters
    ----------
    file_path: str
        Original file path.

    Returns
    -------
    tuple
        name: str
            Name of the file.
        extension: str
            File extension.

    """
    path = Path(file_path)
    name = path.stem
    extension = path.suffix
    return name, extension


def control_warnings() -> None:
    """
    Turn warnings on/off.

        See also: `pelicun/pytest.ini`. Devs: make sure to update that
        file when addressing & eliminating warnings.

    """
    if not sys.warnoptions:
        # Here we specify *specific* warnings to ignore.
        # 'message' -- a regex that the warning message must match

        # Note: we ignore known warnings emitted from our dependencies
        # and plan to address them soon.

        warnings.filterwarnings(
            action='ignore', message='.*Use to_numeric without passing `errors`.*'
        )
        warnings.filterwarnings(
            action='ignore', message=".*errors='ignore' is deprecated.*"
        )
        warnings.filterwarnings(
            action='ignore',
            message='.*The previous implementation of stack is deprecated.*',
        )
        warnings.filterwarnings(
            action='ignore',
            message='.*Setting an item of incompatible dtype is deprecated.*',
        )
        warnings.filterwarnings(
            action='ignore',
            message='.*DataFrame.groupby with axis=1 is deprecated.*',
        )


def load_default_options() -> dict:
    """
    Load the default_config.json file to set options to default values.

    Returns
    -------
    dict
        Default options

    """
    with Path(pelicun_path / 'settings/default_config.json').open(
        encoding='utf-8'
    ) as f:
        default_config = json.load(f)

    return default_config['Options']


def update_vals(
    update_value: dict, primary: dict, update_path: str, primary_path: str
) -> None:
    """
    Transfer values between nested dictionaries.

    Updates the values of the `update` nested dictionary with
    those provided in the `primary` nested dictionary. If a key
    already exists in update, and does not map to another
    dictionary, the value is left unchanged.

    Parameters
    ----------
    update_value: dict
        Dictionary -which can contain nested dictionaries- to be
        updated based on the values of `primary`. New keys existing
        in `primary` are added to `update`. Values of which keys
        already exist in `primary` are left unchanged.
    primary: dict
        Dictionary -which can contain nested dictionaries- to
        be used to update the values of `update`.
    update_path: str
        Identifier for the update dictionary. Used to make error
        messages more meaningful.
    primary_path: str
        Identifier for the update dictionary. Used to make error
        messages more meaningful.

    Raises
    ------
    ValueError
      If primary[key] is dict but update[key] is not.
    ValueError
      If update[key] is dict but primary[key] is not.

    """
    # we go over the keys of `primary`
    for key in primary:  # noqa: PLC0206
        # if `primary[key]` is a dictionary:
        if isinstance(primary[key], dict):
            # if the same `key` does not exist in update,
            # we associate it with an empty dictionary.
            if key not in update_value:
                update_value[key] = {}
            # if it exists already, it should map to
            # a dictionary.
            elif not isinstance(update_value[key], dict):
                msg = (
                    f'{update_path}["{key}"] '
                    'should map to a dictionary. '
                    'The specified value is '
                    f'{update_path}["{key}"] = {update_value[key]}, but '
                    f'the default value is '
                    f'{primary_path}["{key}"] = {primary[key]}. '
                    f'Please revise {update_path}["{key}"].'
                )
                raise ValueError(msg)
            # With both being dictionaries, we use recursion.
            update_vals(
                update_value[key],
                primary[key],
                f'{update_path}["{key}"]',
                f'{primary_path}["{key}"]',
            )
        # if `primary[key]` is NOT a dictionary:
        elif key not in update_value:
            update_value[key] = primary[key]
        elif isinstance(update_value[key], dict):
            msg = (
                f'{update_path}["{key}"] '
                'should not map to a dictionary. '
                f'The specified value is '
                f'{update_path}["{key}"] = {update_value[key]}, but '
                f'the default value is '
                f'{primary_path}["{key}"] = {primary[key]}. '
                f'Please revise {update_path}["{key}"].'
            )
            raise ValueError(msg)


def merge_default_config(user_config: dict | None) -> dict:
    """
    Merge default config with user's options.

    Merge the user-specified config with the configuration defined in
    the default_config.json file. If the user-specified config does
    not include some option available in the default options, then the
    default option is used in the merged config.

    Parameters
    ----------
    user_config: dict
        User-specified configuration dictionary

    Returns
    -------
    dict
        Merged configuration dictionary

    """
    config = user_config  # start from the user's config
    default_config = load_default_options()

    if config is None:
        config = {}

    # We fill out the user's config with the values available in the
    # default config that were not set.
    # We use a recursive function to handle nesting.
    update_vals(config, default_config, 'user_settings', 'default_settings')

    return config


# https://stackoverflow.com/questions/52445559/
# how-can-i-type-hint-a-function-where-the-
# return-type-depends-on-the-input-type-o


@overload
def convert_to_SimpleIndex(
    data: pd.DataFrame, axis: int = 0, *, inplace: bool = False
) -> pd.DataFrame: ...


@overload
def convert_to_SimpleIndex(
    data: pd.Series, axis: int = 0, *, inplace: bool = False
) -> pd.Series: ...


def convert_to_SimpleIndex(  # noqa: N802
    data: pd.DataFrame | pd.Series, axis: int = 0, *, inplace: bool = False
) -> pd.DataFrame | pd.Series:
    """
    Convert the index of a DataFrame to a simple, one-level index.

    The target index uses standard SimCenter convention to identify
    different levels: a dash character ('-') is used to separate each
    level of the index.

    Parameters
    ----------
    data: DataFrame
        The DataFrame that will be modified.
    axis: int, optional, default:0
        Identifies if the index (0) or the columns (1) shall be
        edited.
    inplace: bool, optional, default:False
        If yes, the operation is performed directly on the input
        DataFrame and not on a copy of it.

    Returns
    -------
    DataFrame
        The modified DataFrame

    Raises
    ------
    ValueError
        When an invalid axis parameter is specified

    """
    if axis in {0, 1}:
        data_mod = data if inplace else data.copy()

        if axis == 0:
            # only perform this if there are multiple levels
            if data.index.nlevels > 1:
                simple_name = '-'.join(
                    [n if n is not None else '' for n in data.index.names]
                )
                simple_index = [
                    '-'.join([str(id_i) for id_i in idx]) for idx in data.index
                ]

                data_mod.index = pd.Index(simple_index, name=simple_name)
                data_mod.index.name = simple_name

        elif axis == 1:
            # only perform this if there are multiple levels
            if data.columns.nlevels > 1:
                simple_name = '-'.join(
                    [n if n is not None else '' for n in data.columns.names]
                )
                simple_index = [
                    '-'.join([str(id_i) for id_i in idx]) for idx in data.columns
                ]

                data_mod.columns = pd.Index(simple_index, name=simple_name)
                data_mod.columns.name = simple_name

    else:
        msg = f'Invalid axis parameter: {axis}'
        raise ValueError(msg)

    return data_mod


@overload
def convert_to_MultiIndex(
    data: pd.DataFrame, axis: int = 0, *, inplace: bool = False
) -> pd.DataFrame: ...


@overload
def convert_to_MultiIndex(
    data: pd.Series, axis: int = 0, *, inplace: bool = False
) -> pd.Series: ...


def convert_to_MultiIndex(  # noqa: N802
    data: pd.DataFrame | pd.Series, axis: int = 0, *, inplace: bool = False
) -> pd.DataFrame | pd.Series:
    """
    Convert the index of a DataFrame to a MultiIndex.

    We assume that the index uses standard SimCenter convention to
    identify different levels: a dash character ('-') is expected to
    separate each level of the index.

    Parameters
    ----------
    data: DataFrame
        The DataFrame that will be modified.
    axis: int, optional, default:0
        Identifies if the index (0) or the columns (1) shall be
        edited.
    inplace: bool, optional, default:False
        If yes, the operation is performed directly on the input
        DataFrame and not on a copy of it.

    Returns
    -------
    DataFrame
        The modified DataFrame.

    Raises
    ------
    ValueError
        If an invalid axis is specified.

    """
    # check if the requested axis is already a MultiIndex
    if ((axis == 0) and (isinstance(data.index, pd.MultiIndex))) or (
        (axis == 1) and (isinstance(data.columns, pd.MultiIndex))
    ):
        # if yes, return the data unchanged
        return data

    if axis == 0:
        index_labels = [str(label).split('-') for label in data.index]

    elif axis == 1:
        index_labels = [str(label).split('-') for label in data.columns]

    else:
        msg = f'Invalid axis parameter: {axis}'
        raise ValueError(msg)

    max_lbl_len = np.max([len(labels) for labels in index_labels])

    for l_i, labels in enumerate(index_labels):
        if len(labels) != max_lbl_len:
            labels += [''] * (max_lbl_len - len(labels))  # noqa: PLW2901
            index_labels[l_i] = labels

    index_labels_np = np.array(index_labels)

    if index_labels_np.shape[1] > 1:
        data_mod = data if inplace else data.copy()

        if axis == 0:
            data_mod.index = pd.MultiIndex.from_arrays(index_labels_np.T)

        else:
            data_mod.columns = pd.MultiIndex.from_arrays(index_labels_np.T)

        return data_mod

    return data


def convert_dtypes(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Convert columns to a numeric datatype whenever possible.

    The function replaces None with NA otherwise columns containing
    None would continue to have the `object` type.

    Parameters
    ----------
    dataframe: DataFrame
        The DataFrame that will be modified.

    Returns
    -------
    DataFrame
        The modified DataFrame.

    """
    with (
        pd.option_context('future.no_silent_downcasting', True),  # noqa: FBT003
        pd.option_context('mode.copy_on_write', True),  # noqa: FBT003
    ):
        dataframe = dataframe.fillna(value=np.nan).infer_objects()
    # note: `axis=0` applies the function to the columns
    # note: ignoring errors is a bad idea and should never be done. In
    # this case, however, that's not what we do, despite the name of
    # this parameter. We simply don't convert the dtype of columns
    # that cannot be interpreted as numeric. That's what
    # `errors='ignore'` does.
    # See:
    # https://pandas.pydata.org/docs/reference/api/pandas.to_numeric.html
    return dataframe.apply(
        lambda x: pd.to_numeric(x, errors='ignore'),  # type:ignore
        axis=0,
    )


def show_matrix(
    data: np.ndarray | pd.DataFrame, *, use_describe: bool = False
) -> None:
    """
    Print a matrix in a nice way using a DataFrame.

    Parameters
    ----------
    data: array-like
        The matrix data to display. Can be any array-like structure
        that pandas can convert to a DataFrame.
    use_describe: bool, default: False
        If True, provides a descriptive statistical summary of the
        matrix including specified percentiles.
        If False, simply prints the matrix as is.

    """
    if use_describe:
        pp.pprint(
            pd.DataFrame(data).describe(percentiles=[0.01, 0.1, 0.5, 0.9, 0.99])
        )
    else:
        pp.pprint(pd.DataFrame(data))


def multiply_factor_multiple_levels(
    df: pd.DataFrame,
    conditions: dict,
    factor: float,
    axis: int = 0,
    *,
    raise_missing: bool = True,
) -> None:
    """
    Multiply a value to selected rows, in place.

    Multiplies a value to selected rows of a DataFrame that is indexed
    with a hierarchical index (pd.MultiIndex). The change is done in
    place.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame to be modified.
    conditions: dict
        A dictionary mapping level names with a single value. Only the
        rows where the index levels have the provided values will be
        affected. The dictionary can be empty, in which case all rows
        will be affected, or contain only some levels and values, in
        which case only the matching rows will be affected.
    factor: float
        Scaling factor to use.
    axis: int
        With 0 the condition is checked against the DataFrame's index,
        otherwise with 1 it is checked against the DataFrame's
        columns.
    raise_missing: bool
        Raise an error if no rows are matching the given conditions.

    Raises
    ------
    ValueError
        If the provided `axis` values is not either 0 or 1.
    ValueError
        If there are no rows matching the conditions and raise_missing
        is True.

    """
    if axis == 0:
        idx_to_use = df.index
    elif axis == 1:
        idx_to_use = df.columns
    else:
        msg = f'Invalid axis: `{axis}`'
        raise ValueError(msg)

    mask = pd.Series(data=True, index=idx_to_use)

    # Apply each condition to update the mask
    for level, value in conditions.items():
        mask &= idx_to_use.get_level_values(level) == value

    if np.all(mask == False) and raise_missing:  # noqa: E712
        msg = f'No rows found matching the conditions: `{conditions}`'
        raise ValueError(msg)

    if axis == 0:
        df.iloc[mask.to_numpy()] *= factor
    else:
        df.iloc[:, mask.to_numpy()] *= factor


def _warning(
    message: str,
    category: type[Warning],
    filename: str,
    lineno: int,
    file: Any = None,  # noqa: ARG001, ANN401
    line: Any = None,  # noqa: ARG001, ANN401
) -> None:
    """
    Display warnings in a custom format.

    Custom warning function to format and print warnings more
    attractively. This function modifies how warning messages are
    displayed, emphasizing the file path and line number from where
    the warning originated.

    Parameters
    ----------
    message: str
        The warning message to be displayed.
    category: Warning
        The category of the warning (unused, but required for
        compatibility with standard warning signature).
    filename: str
        The path of the file from which the warning is issued. The
        function simplifies the path for display.
    lineno: int
        The line number in the file at which the warning is issued.
    file: file-like object, optional
        The target file object to write the warning to (unused, but
        required for compatibility with standard warning signature).
    line: str, optional
        Line of code causing the warning (unused, but required for
        compatibility with standard warning signature).

    """
    # pylint:disable = unused-argument
    if category != PelicunWarning:
        if '\\' in filename:
            file_path = filename.split('\\')
        elif '/' in filename:
            file_path = filename.split('/')
        else:
            file_path = None

        python_file = '/'.join(file_path[-3:]) if file_path is not None else filename
        print(f'WARNING in {python_file} at line {lineno}\n{message}\n')  # noqa: T201
    else:
        print(message)  # noqa: T201


warnings.showwarning = _warning  # type: ignore


def describe(
    data: pd.DataFrame | pd.Series | np.ndarray,
    percentiles: tuple[float, ...] = (
        0.001,
        0.023,
        0.10,
        0.159,
        0.5,
        0.841,
        0.90,
        0.977,
        0.999,
    ),
) -> pd.DataFrame:
    """
    Extend descriptive statistics.

    Provides extended descriptive statistics for given data, including
    percentiles and log standard deviation for applicable columns.
    This function accepts both pandas Series and DataFrame objects
    directly, or any array-like structure which can be converted to
    them. It calculates common descriptive statistics and optionally
    adds log standard deviation for columns where all values are
    positive.

    Parameters
    ----------
    data: pd.Series, pd.DataFrame, or array-like
        The data to describe. If array-like, it is converted to a
        DataFrame or Series before analysis.
    percentiles: tuple of float, optional
        Specific percentiles to include in the output. Default
        includes an extensive range tailored to provide a detailed
        summary.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the descriptive statistics of the input
        data, transposed so that each descriptive statistic is a row.

    """
    if isinstance(data, np.ndarray):
        vals = data

        if vals.ndim == 1:
            data = pd.Series(vals, name=0)
        else:
            cols = np.arange(vals.shape[1])
            data = pd.DataFrame(vals, columns=cols)

    # convert Series to a DataFrame
    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)

    desc = pd.DataFrame(data.describe(list(percentiles)).T)

    # add log standard deviation to the stats
    desc.insert(3, 'log_std', np.nan)
    desc = desc.T

    for col in desc.columns:
        if np.min(data[col]) > 0.0:
            desc.loc['log_std', col] = np.std(np.log(data[col]), ddof=1)

    return desc


def str2bool(v: str | bool) -> bool:  # noqa: FBT001
    """
    Convert a string representation of truth to boolean True or False.

    This function is designed to convert string inputs that represent
    boolean values into actual Python boolean types. It handles
    typical representations of truthiness and falsiness, and is case
    insensitive.

    Parameters
    ----------
    v: str or bool
        The value to convert into a boolean. This can be a boolean
        itself (in which case it is simply returned) or a string that
        is expected to represent a boolean value.

    Returns
    -------
    bool
        The boolean value corresponding to the input.

    Raises
    ------
    argparse.ArgumentTypeError
        If `v` is a string that does not correspond to a boolean
        value, an error is raised indicating that a boolean value was
        expected.

    """
    # courtesy of Maxim @ Stackoverflow

    if isinstance(v, bool):
        return v
    if v.lower() in {'yes', 'true', 'True', 't', 'y', '1'}:
        return True
    if v.lower() in {'no', 'false', 'False', 'f', 'n', '0'}:
        return False
    msg = 'Boolean value expected.'
    raise argparse.ArgumentTypeError(msg)


def float_or_None(string: str) -> float | None:  # noqa: N802
    """
    Convert strings to float or None.

    Parameters
    ----------
    string: str
        A string

    Returns
    -------
    float or None
        A float, if the given string can be converted to a
        float. Otherwise, it returns None

    """
    try:
        return float(string)
    except ValueError:
        return None


def int_or_None(string: str) -> int | None:  # noqa: N802
    """
    Convert strings to int or None.

    Parameters
    ----------
    string: str
        A string

    Returns
    -------
    int or None
        An int, if the given string can be converted to an
        int. Otherwise, it returns None

    """
    try:
        return int(string)
    except ValueError:
        return None


def check_if_str_is_na(string: Any) -> bool:  # noqa: ANN401
    """
    Check if the provided string can be interpreted as N/A.

    Parameters
    ----------
    string: object
            The string to evaluate

    Returns
    -------
    bool
        The evaluation result. Yes, if the string is considered N/A.
    """
    na_vals = {
        '',
        'N/A',
        '-1.#QNAN',
        'null',
        'None',
        '<NA>',
        'nan',
        '-NaN',
        '1.#IND',
        'NaN',
        '#NA',
        '1.#QNAN',
        'NULL',
        '-nan',
        '#N/A',
        '#N/A N/A',
        'n/a',
        '-1.#IND',
        'NA',
    }
    # obtained from Pandas' internal STR_NA_VALUES variable.

    return isinstance(string, str) and string in na_vals


def with_parsed_str_na_values(df: pd.DataFrame) -> pd.DataFrame:
    """
    Identify string values interpretable as N/A.

    Given a dataframe, this function identifies values that have
    string type and can be interpreted as N/A, and replaces them with
    actual NA's.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to process

    Returns
    -------
    pd.DataFrame
        The dataframe with proper N/A values.
    """
    # Replace string NA values with actual NaNs
    return df.apply(
        lambda col: col.map(lambda x: np.nan if check_if_str_is_na(x) else x)
    )


def dedupe_index(dataframe: pd.DataFrame, dtype: type = str) -> pd.DataFrame:
    """
    Add a `uid` level to the index.

    Modifies the index of a DataFrame to ensure all index elements are
    unique by adding an extra level.  Assumes that the DataFrame's
    original index is a MultiIndex with specified names. A unique
    identifier ('uid') is added as an additional index level based on
    the cumulative count of occurrences of the original index
    combinations.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The DataFrame whose index is to be modified. It must have a
        MultiIndex.
    dtype: type, optional
        The data type for the new index level 'uid'. Defaults to str.

    Returns
    -------
    dataframe: pd.DataFrame
        The original dataframe with an additional `uid` level at the
        index.

    """
    inames = dataframe.index.names
    dataframe = dataframe.reset_index()
    dataframe['uid'] = (dataframe.groupby([*inames]).cumcount()).astype(dtype)
    dataframe = dataframe.set_index([*inames, 'uid'])
    return dataframe.sort_index()


# Input specs

EDP_to_demand_type = {
    # Drifts
    'Story Drift Ratio': 'PID',
    'Peak Interstory Drift Ratio': 'PID',
    'Roof Drift Ratio': 'PRD',
    'Peak Roof Drift Ratio': 'PRD',
    'Damageable Wall Drift': 'DWD',
    'Racking Drift Ratio': 'RDR',
    'Mega Drift Ratio': 'PMD',
    'Residual Drift Ratio': 'RID',
    'Residual Interstory Drift Ratio': 'RID',
    'Peak Effective Drift Ratio': 'EDR',
    # Floor response
    'Peak Floor Acceleration': 'PFA',
    'Peak Floor Velocity': 'PFV',
    'Peak Floor Displacement': 'PFD',
    # Component response
    'Peak Link Rotation Angle': 'LR',
    'Peak Link Beam Chord Rotation': 'LBR',
    # Wind Intensity
    'Peak Gust Wind Speed': 'PWS',
    # Wind Demands
    'Peak Wind Force': 'PWF',
    'Peak Internal Force': 'PIF',
    'Peak Line Force': 'PLF',
    'Peak Wind Pressure': 'PWP',
    # Inundation Intensity
    'Peak Inundation Height': 'PIH',
    # Shaking Intensity
    'Peak Ground Acceleration': 'PGA',
    'Peak Ground Velocity': 'PGV',
    'Spectral Acceleration': 'SA',
    'Spectral Velocity': 'SV',
    'Spectral Displacement': 'SD',
    'Peak Spectral Acceleration': 'SA',
    'Peak Spectral Velocity': 'SV',
    'Peak Spectral Displacement': 'SD',
    'Permanent Ground Deformation': 'PGD',
    # Placeholder for advanced calculations
    'One': 'ONE',
}


def dict_raise_on_duplicates(ordered_pairs: list[tuple]) -> dict:
    """
    Construct a dictionary from a list of key-value pairs.

    Constructs a dictionary from a list of key-value pairs, raising an
    exception if duplicate keys are found.
    This function ensures that no two pairs have the same key. It is
    particularly useful when parsing JSON-like data where unique keys
    are expected but not enforced by standard parsing methods.

    Parameters
    ----------
    ordered_pairs: list of tuples
        A list of tuples, each containing a key and a value. Keys are
        expected to be unique across the list.

    Returns
    -------
    dict
        A dictionary constructed from the ordered_pairs without any
        duplicates.

    Raises
    ------
    ValueError
        If a duplicate key is found in the input list, a ValueError is
        raised with a message indicating the duplicate key.

    Examples
    --------
    >>> dict_raise_on_duplicates(
    ...     [("key1", "value1"), ("key2", "value2"), ("key1", "value3")]
    ... )
    ValueError: duplicate key: key1

    Notes
    -----
    This implementation is useful for contexts in which data integrity
    is crucial and key uniqueness must be ensured.

    """
    d = {}
    for k, v in ordered_pairs:
        if k in d:
            msg = f'duplicate key: {k}'
            raise ValueError(msg)
        d[k] = v
    return d


def parse_units(  # noqa: C901
    custom_file: str | None = None, *, preserve_categories: bool = False
) -> dict:
    """
    Parse the unit conversion factor JSON file and return a dictionary.

    Parameters
    ----------
    custom_file: str, optional
        If a custom file is provided, only the units specified in the
        custom file are used.
    preserve_categories: bool, optional
        If True, maintains the original data types of category
        values from the JSON file. If False, converts all values
        to floats and flattens the dictionary structure, ensuring
        that each unit name is globally unique across categories.


    Returns
    -------
    dict
        A dictionary where keys are unit names and values are
        their corresponding conversion factors. If
        `preserve_categories` is True, the dictionary may maintain
        its original nested structure based on the JSON file. If
        `preserve_categories` is False, the dictionary is flattened
        to have globally unique unit names.

    """

    def get_contents(file_path: Path, *, preserve_categories: bool = False) -> dict:  # noqa: C901
        """
        Map unit names to conversion factors.

        Parses a unit conversion factors JSON file and returns a
        dictionary mapping unit names to conversion factors.

        This function allows the use of a custom JSON file for
        defining unit conversion factors or defaults to a predefined
        file. It ensures that each unit name is unique and that all
        conversion factors are float values. Additionally, it supports
        the option to preserve the original data types of category
        values from the JSON.

        Parameters
        ----------
        file_path: str
            The file path to a JSON file containing unit conversion
            factors. If not provided, a default file is used.
        preserve_categories: bool, optional
            If True, maintains the original data types of category
            values from the JSON file. If False, converts all values
            to floats and flattens the dictionary structure, ensuring
            that each unit name is globally unique across categories.

        Returns
        -------
        dict
            A dictionary where keys are unit names and values are
            their corresponding conversion factors. If
            `preserve_categories` is True, the dictionary may maintain
            its original nested structure based on the JSON file.

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist.
        ValueError
            If a unit name is duplicated or other JSON structure issues are present.
        TypeError
            If a conversion factor is not a float.
        TypeError
            If any value that needs to be converted to float cannot be
            converted.

        Examples
        --------
        >>> parse_units('custom_units.json')
        { 'm': 1.0, 'cm': 0.01, 'mm': 0.001 }

        >>> parse_units('custom_units.json', preserve_categories=True)
        { 'Length': {'m': 1.0, 'cm': 0.01, 'mm': 0.001} }

        """
        try:
            with Path(file_path).open(encoding='utf-8') as f:
                dictionary = json.load(f, object_pairs_hook=dict_raise_on_duplicates)
        except FileNotFoundError as exc:
            msg = f'{file_path} was not found.'
            raise FileNotFoundError(msg) from exc
        except json.decoder.JSONDecodeError as exc:
            msg = f'{file_path} is not a valid JSON file.'
            raise ValueError(msg) from exc
        for category_dict in list(dictionary.values()):
            # ensure all first-level keys point to a dictionary
            if not isinstance(category_dict, dict):
                msg = (
                    f'{file_path} contains first-level keys '
                    "that don't point to a dictionary"
                )
                raise TypeError(msg)
            # convert values to float
            try:
                for key, val in category_dict.items():
                    category_dict[key] = float(val)
            except (ValueError, TypeError) as exc:
                msg = (
                    f'Unit {key} has a value of {val} '
                    'which cannot be interpreted as a float'
                )
                raise type(exc)(msg) from exc

        if preserve_categories:
            return dictionary

        flattened = {}
        for category in dictionary:
            for unit_name, factor in dictionary[category].items():
                if unit_name in flattened:
                    msg = f'{unit_name} defined twice in {file_path}.'
                    raise ValueError(msg)
                flattened[unit_name] = factor

        return flattened

    if custom_file:
        return get_contents(
            Path(custom_file), preserve_categories=preserve_categories
        )

    return get_contents(
        pelicun_path / 'settings/default_units.json',
        preserve_categories=preserve_categories,
    )


def convert_units(  # noqa: C901
    values: float | list[float] | np.ndarray,
    unit: str,
    to_unit: str,
    category: str | None = None,
) -> float | list[float] | np.ndarray:
    """
    Convert numeric values between different units.

    Supports conversion within a specified category of units and
    automatically infers the category if not explicitly provided. It
    maintains the type of the input in the output.

    Parameters
    ----------
    values: (float | list[float] | np.ndarray)
      The numeric value(s) to convert.
    unit: (str)
      The current unit of the values.
    to_unit: (str)
      The target unit to convert the values into.
    category: (Optional[str])
      The category of the units (e.g., 'length', 'pressure'). If not
      provided, the category will be inferred based on the provided
      units.

    Returns
    -------
    float or list[float] or np.ndarray
      The converted value(s) in the target unit, in the same data type
      as the input values.

    Raises
    ------
    TypeError
      If the input `values` are not of type float, list, or
      np.ndarray.
    ValueError
      If the `unit`, `to_unit`, or `category` is unknown or if `unit`
      and `to_unit` are not in the same category.

    """
    if isinstance(values, (float, list)):
        vals = np.atleast_1d(values)
    elif isinstance(values, np.ndarray):
        vals = values
    else:
        msg = 'Invalid input type for `values`'
        raise TypeError(msg)

    # load default units
    all_units = parse_units(preserve_categories=True)

    # if a category is given use it, otherwise try to determine it
    if category:
        if category not in all_units:
            msg = f'Unknown category: `{category}`'
            raise ValueError(msg)
        units = all_units[category]
        for unt in unit, to_unit:
            if unt not in units:
                msg = f'Unknown unit: `{unt}`'
                raise ValueError(msg)
    else:
        unit_category: str | None = None
        for key in all_units:
            units = all_units[key]
            if unit in units:
                unit_category = key
                break
        if not unit_category:
            msg = f'Unknown unit `{unit}`'
            raise ValueError(msg)
        units = all_units[unit_category]
        if to_unit not in units:
            msg = (
                f'`{unit}` is a `{unit_category}` unit, but `{to_unit}` '
                f'is not specified in that category.'
            )
            raise ValueError(msg)

    # convert units
    from_factor = units[unit]
    to_factor = units[to_unit]
    new_values = vals * float(from_factor) / float(to_factor)

    # return the results in the same type as that of the provided
    # values
    if isinstance(values, float):
        return new_values[0]
    if isinstance(values, list):
        return new_values.tolist()
    return new_values


def stringterpolation(
    arguments: str,
) -> Callable[[np.ndarray], np.ndarray]:
    """
    Linear interpolation from strings.

    Turns a string of specially formatted arguments into a multilinear
    interpolating function.

    Parameters
    ----------
    arguments: str
        String of arguments containing Y values and X values,
        separated by a pipe symbol (`|`). Individual values are
        separated by commas (`,`). Example:
        arguments = 'y1,y2,y3|x1,x2,x3'

    Returns
    -------
    Callable
        A callable interpolating function

    """
    split = arguments.split('|')
    x_vals = split[1].split(',')
    y_vals = split[0].split(',')
    x = np.array(x_vals, dtype=float)
    y = np.array(y_vals, dtype=float)

    return interp1d(x=x, y=y, kind='linear')


def invert_mapping(original_dict: dict) -> dict:
    """
    Inverts a dictionary mapping from key to list of values.

    Parameters
    ----------
    original_dict: dict
        Dictionary with values that are lists of hashable items.

    Returns
    -------
    dict
        New dictionary where each item in the original value lists
        becomes a key and the original key becomes the corresponding
        value.

    Raises
    ------
    ValueError
        If any value in the original dictionary's value lists appears
        more than once.

    """
    inverted_dict = {}
    for key, value_list in original_dict.items():
        for value in value_list:
            if value in inverted_dict:
                msg = 'Cannot invert mapping with duplicate values.'
                raise ValueError(msg)
            inverted_dict[value] = key
    return inverted_dict


def get(
    d: dict | None,
    path: str,
    default: Any | None = None,  # noqa: ANN401
) -> Any:  # noqa: ANN401
    """
    Path-like dictionary value retrieval.

    Retrieves a value from a nested dictionary using a path with '/'
    as the separator.

    Parameters
    ----------
    d: dict
        The dictionary to search.
    path: str
        The path to the desired value, with keys separated by '/'.
    default: Any, optional
        The value to return if the path is not found. Defaults to
        None.

    Returns
    -------
    Any
        The value found at the specified path, or the default value if
        the path is not found.

    Examples
    --------
    >>> config = {
    ...     "DL": {
    ...         "Outputs": {
    ...             "Format": {
    ...                 "JSON": "desired_value"
    ...             }
    ...         }
    ...     }
    ... }
    >>> get(config, '/DL/Outputs/Format/JSON', default='default_value')
    'desired_value'
    >>> get(config, '/DL/Outputs/Format/XML', default='default_value')
    'default_value'

    """
    if d is None:
        return default
    keys = path.strip('/').split('/')
    current_dict = d
    try:
        for key in keys:
            current_dict = current_dict[key]
        return current_dict  # noqa: TRY300
    except (KeyError, TypeError):
        return default


def update(
    d: dict[str, Any],
    path: str,
    value: Any,  # noqa: ANN401
    *,
    only_if_empty_or_none: bool = False,
) -> None:
    """
    Set a value in a nested dictionary using a path with '/' as the separator.

    Parameters
    ----------
    d: dict
        The dictionary to update.
    path: str
        The path to the desired value, with keys separated by '/'.
    value: Any
        The value to set at the specified path.
    only_if_empty_or_none: bool, optional
        If True, only update the value if it is None or an empty
        dictionary. Defaults to False.

    Examples
    --------
    >>> d = {}
    >>> update(d, 'x/y/z', 1)
    >>> d
    {'x': {'y': {'z': 1}}}

    >>> update(d, 'x/y/z', 2, only_if_empty_or_none=True)
    >>> d
    {'x': {'y': {'z': 1}}}  # value remains 1 since it is not empty or None

    >>> update(d, 'x/y/z', 2)
    >>> d
    {'x': {'y': {'z': 2}}}  # value is updated to 2

    """
    keys = path.strip('/').split('/')
    current_dict = d
    for key in keys[:-1]:
        if key not in current_dict or not isinstance(current_dict[key], dict):
            current_dict[key] = {}
        current_dict = current_dict[key]
    if only_if_empty_or_none:
        if is_unspecified(current_dict, keys[-1]):
            current_dict[keys[-1]] = value
    else:
        current_dict[keys[-1]] = value


def is_unspecified(d: dict[str, Any], path: str) -> bool:
    """
    Check if something is specified.

    Checks if a value in a nested dictionary is either non-existent,
    None, NaN, or an empty dictionary or list.

    Parameters
    ----------
    d: dict
        The dictionary to search.
    path: str
        The path to the desired value, with keys separated by '/'.

    Returns
    -------
    bool
        True if the value is non-existent, None, or an empty
        dictionary or list. False otherwise.

    Examples
    --------
    >>> config = {
    ...     "DL": {
    ...         "Outputs": {
    ...             "Format": {
    ...                 "JSON": "desired_value",
    ...                 "EmptyDict": {}
    ...             }
    ...         }
    ...     }
    ... }
    >>> is_unspecified(config, '/DL/Outputs/Format/JSON')
    False
    >>> is_unspecified(config, '/DL/Outputs/Format/XML')
    True
    >>> is_unspecified(config, '/DL/Outputs/Format/EmptyDict')
    True

    """
    value = get(d, path, default=None)
    if value is None:
        return True
    if pd.isna(value):
        return True
    if value == {}:
        return True
    return value == []


def is_specified(d: dict[str, Any], path: str) -> bool:
    """
    Opposite of `is_unspecified()`.

    Parameters
    ----------
    d: dict
        The dictionary to search.
    path: str
        The path to the desired value, with keys separated by '/'.

    Returns
    -------
    bool
        True if the value is specified, False otherwise.

    """
    return not is_unspecified(d, path)


def ensure_value(value: T | None) -> T:
    """
    Ensure a variable is not None.

    This function checks that the provided variable is not None. It is
    used to assist with type hinting by avoiding repetitive `assert
    value is not None` statements throughout the code.

    Parameters
    ----------
    value : Optional[T]
        The variable to check, which can be of any type or None.

    Returns
    -------
    T
        The same variable, guaranteed to be non-None.

    Raises
    ------
    TypeError
        If the provided variable is None.

    """
    if value is None:
        raise TypeError
    return value
